"""
Test patch utilities.
"""
import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.exceptions import PatchAttributeError, PatchDimError
from dascore.utils.patch import _AttrsCoordsMixer, _merge_patches


@dc.patch_function(required_dims=("time", "distance"))
def time_dist_func(patch):
    """A dummy function to test time, distance coord requirement."""
    return patch


@dc.patch_function(required_attrs=("bob",))
def require_bob(patch):
    """Require bob attribute"""
    return patch


@dc.patch_function(required_attrs={"bob": "what about?"})
def require_bob_value(patch):
    """Require bob attribute with specific value."""
    return patch


class TestPatchFunction:
    """Tests for patch function decorator."""

    def test_no_call(self, random_patch):
        """Ensure the decorator can be used with or without parens"""

        @dc.patch_function
        def func1(patch):
            """first test func"""
            return patch

        @dc.patch_function()
        def func2(patch):
            """second test func"""
            return patch

        pa1 = func1(random_patch)
        pa2 = func2(random_patch)
        assert pa1.equals(pa2)

    def test_has_required_dims(self, random_patch):
        """Test that the required dimension check works."""
        # passes if this doesn't raise
        time_dist_func(random_patch)

    def test_doesnt_have_required_dims(self, random_patch):
        """Raises if patch doesn't have required dimensions."""
        pa = random_patch.rename(time="for_a_hug")

        with pytest.raises(PatchDimError):
            time_dist_func(pa)

    def test_required_attrs(self, random_patch):
        """Tests for requiring certain attrs."""
        with pytest.raises(PatchAttributeError):
            require_bob(random_patch)

    def test_require_attr_exists(self, random_patch):
        """ "test for requiring an attr exists but not checking value"""
        pa1 = random_patch.update_attrs(bob=1)
        pa2 = pa1.update_attrs(bob=2)
        # both these should not raise
        require_bob(pa1)
        require_bob(pa2)

    def test_require_attr_and_value(self, random_patch):
        """Tests for requiring attribute and specific value."""
        pa1 = random_patch.update_attrs(bob="what about?")
        pa2 = random_patch.update_attrs(bob="lightening")
        require_bob_value(pa1)  # should not raise
        with pytest.raises(PatchAttributeError, match="lightening"):
            require_bob_value(pa2)

    def test_access_original_function(self, random_patch):
        """Ensure the original function is still accessible."""
        hist1 = ",".join(random_patch.attrs["history"])
        # this calls the original function so it should bypass history logging
        out = require_bob.func(random_patch)
        hist2 = ",".join(out.attrs["history"])
        assert hist1 == hist2


class TestHistory:
    """Tests for tracking patch processing history."""

    def test_simple_history(self, random_patch):
        """just make sure function is logged."""
        out = time_dist_func(random_patch)
        history = out.attrs["history"]
        assert len(history) == 1
        assert "time_dist_func" in history[0]

    def test_original_history_unchanged(self, random_patch):
        """Ensure logging history only occurs on new Patch"""
        first_history = list(random_patch.attrs["history"])
        _ = time_dist_func(random_patch)
        last_history = list(random_patch.attrs["history"])
        assert first_history == last_history


class TestAttrsCoordsMixer:
    """Tests for handling complex interaction between attrs and coords."""

    @pytest.fixture()
    def attrs(self, random_patch):
        """return an attrs from a patch"""
        return random_patch.attrs

    @pytest.fixture()
    def coords(self, random_patch):
        """return an attrs from a patch"""
        return random_patch.coords

    @pytest.fixture()
    def mixer(self, attrs, coords, random_patch):
        """Return a mixer instance."""
        return _AttrsCoordsMixer(attrs, coords, random_patch.dims)

    def test_original_attrs_unchanged(self, mixer, attrs):
        """ensure the original attrs don't change."""
        t1 = attrs["time_min"]
        td = np.timedelta64(1, "s")
        mixer.update_attrs(time_min=t1 + td)
        new_attr, _ = mixer()
        assert new_attr is not attrs
        assert attrs["time_min"] + td == new_attr["time_min"]

    def test_original_coords_unchanged(self, mixer, coords, attrs):
        """ensure the original coords don't change."""
        t1 = attrs["time_min"]
        td = np.timedelta64(10, "s")
        mixer.update_attrs(time_min=t1 + td)
        _, new_coords = mixer()
        assert new_coords is not coords
        np.all(np.equal(coords["time"] + td, new_coords["time"]))

    def test_coords_unchanged(self, coords, attrs, random_patch):
        """ensure the original coords don't change."""
        # this was added to track down some mutation issues
        assert coords["time"].min() == attrs["time_min"]

    def test_starttime_updates_endtime(self, mixer, attrs, coords):
        """Ensure the end time gets updated when setting time_min"""
        t1 = attrs["time_min"]
        t_new = t1 + np.timedelta64(10_000_000, "s")
        mixer.update_attrs(time_min=t_new)
        attr, coords = mixer()
        assert attr["time_min"] == t_new
        # make sure time min was updated in coords
        time = coords["time"]
        assert np.min(time) == t_new

    def test_endtime_updates_starttime(self, mixer, attrs):
        """Ensure the start time gets updated when setting time_max."""
        tdist1 = attrs["time_max"] - attrs["time_min"]
        t2 = attrs["time_max"]
        t_new = t2 - np.timedelta64(10_000_000, "s")
        mixer.update_attrs(time_max=t_new)
        attr, coords = mixer()
        assert attr["time_max"] == t_new
        # make sure time min was updated in coords
        time = coords["time"]
        assert np.max(time) == t_new
        # ensure distance between start/end is the same
        tdist2 = attr["time_max"] - attr["time_min"]
        assert tdist2 == tdist1

    def test_coords_updates_times_and_distance(self, mixer, coords, attrs):
        """
        Ensure updating coords also updates attributes in attrs.
        """
        td = np.timedelta64(10, "s")
        dx = 10
        new_coords_kwarg = {
            "time": coords["time"] + td,
            "distance": coords["distance"] + dx,
        }

        mixer.update_coords(**new_coords_kwarg)
        new_attrs, new_coords = mixer()
        # first ensure coords actually updated
        assert np.all(new_coords["time"] == new_coords_kwarg["time"])
        assert np.all(new_coords["distance"] == new_coords_kwarg["distance"])
        # check attrs time are updated
        assert attrs["time_min"] + td == new_attrs["time_min"]
        assert attrs["time_max"] + td == new_attrs["time_max"]
        # check distance
        assert attrs["distance_min"] + dx == new_attrs["distance_min"]
        assert attrs["distance_max"] + dx == new_attrs["distance_max"]

    def test_relative_time_coord_no_absolute_time(self):
        """Ensure a time axis is still obtained with relative time and no start"""
        coords = dict(
            time=np.arange(10) * 0.1,
            distance=np.arange(10) * 10,
        )
        attrs = {}
        mixer = _AttrsCoordsMixer(attrs, coords, ("time", "distance"))
        new_attrs, new_coords = mixer()
        assert not np.any(pd.isnull(new_coords["time"]))

    def test_update_startttime_string(self, mixer, coords, attrs):
        """check start/end time relationship when starttime is a string."""
        new_start = np.datetime64("2000-01-01")
        duration = attrs["time_max"] - attrs["time_min"]
        mixer.update_attrs(time_min=str(new_start))
        new_attrs, _ = mixer()
        assert new_attrs["time_min"] == new_start
        assert new_attrs["time_max"] == new_start + duration

    def test_update_time_delta(self, mixer, coords, attrs):
        """Updating the time delta should also update endtimes."""
        one_sec = np.timedelta64(1, "s")
        td_old = attrs["d_time"]
        td_new = td_old * 2
        mixer.update_attrs(d_time=td_new)
        new_attrs, new_coords = mixer()
        assert new_attrs["d_time"] == td_new
        # first ensure new time coords are approximately new time delta
        new_time = new_coords["time"]
        tdiff = (new_time[1:] - new_time[:-1]) / one_sec
        assert np.allclose(tdiff, td_new / one_sec)
        # also ensure the endtime has increased proportionately.
        old_duration = attrs["time_max"] - attrs["time_min"]
        new_duration = new_attrs["time_max"] - new_attrs["time_min"]
        assert np.isclose(old_duration / new_duration, td_old / td_new)


class TestMergePatches:
    """Tests for merging patches together."""

    def test_merge_dir_spool(self, adjacent_spool_directory):
        """Ensure merge_patches works on a directory spool."""
        spool = dc.spool(adjacent_spool_directory)
        out = _merge_patches(spool)
        assert len(out) == 1

    def test_deprecated(self, random_patch):
        """Ensure deprecation warning is raised."""
        from dascore.utils.patch import merge_patches

        with pytest.warns(DeprecationWarning, match="merge_patches is deprecated"):
            merge_patches(random_patch)



"""
A spectrogram visualization
"""
from typing import Optional

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import spectrogram as scipy_spectrogram

from dascore.constants import PatchType
from dascore.utils.patch import patch_function
from dascore.utils.plotting import _get_ax, _get_cmap


def _get_1d_array_dim_name(patch):
    """
    Return the flattened (1D) array and dim name.
    Raises a Value Error array is higher than 1D.
    """
    data = patch.data
    if len(data.shape) == 1:  # if array only has one dimension.
        return data, patch.dims[0]
    shape = np.array(data.shape)
    not_one_len = shape > 1
    if np.sum(not_one_len) != 1:
        msg = f"Spectrogram requires 1D patch, not {data.shape}"
        raise ValueError(msg)
    name = patch.dims[np.argmax(not_one_len)]
    out = data.flatten()
    return out, name


@patch_function()
def spectrogram(
    patch: PatchType,
    ax: Optional[plt.Axes] = None,
    cmap="bwr",
    log=False,
    show=False,
) -> plt.Axes:
    """
    Plot a spectrogram of a patch along a collapsed dimension.

    Parameters
    ----------
    ax
        If not None, an axis on which to plot.
    cmap
        A matplotlob color map or code.
    show
        If True call plt.show() else just return axis.
    """
    data, name = _get_1d_array_dim_name(patch)
    d_dim = patch.attrs[f"d_{name}"]
    if name == "time":
        d_dim = d_dim / np.timedelta64(1, "s")
    freqs, dim_x, spec = scipy_spectrogram(data, 1 / d_dim)
    ax = _get_ax(ax)
    norm_class = colors.LogNorm if log else colors.Normalize
    norm = norm_class(vmin=np.min(spec), vmax=np.max(spec))
    cmap = _get_cmap(cmap)
    ax.pcolormesh(dim_x, freqs, spec, shading="gouraud", cmap=cmap, norm=norm)
    ax.set_ylabel(f"Frequency ({name}) [Hz]")
    ax.set_xlabel(f"{name.capitalize()}")
    if name == "time":
        time_str = str(patch.attrs["time_min"]).split(".")[0]
        ax.set_xlabel(f"{name.capitalize()} from {time_str}")
    if show:
        plt.show()
    return ax



"""
Tests for coordinate utilities.
"""
import numpy as np
import pytest

from dascore.utils.coords import Coords
from dascore.utils.misc import register_func


class TestCoords:
    """Tests for DASCore's custom coord class."""

    coord_fixtures = []
    time = [1, 2, 3]
    distance = [1, 2, 3, 4, 5]
    shape = (len(time), len(distance))
    dims = ("time", "distance")

    @pytest.fixture(scope="class")
    @register_func(coord_fixtures)
    def simple_dict_coords(self):
        """Create a simple version of coords from dict."""
        cdict = {"time": self.time, "distance": self.distance}
        return Coords(cdict, self.dims)

    @pytest.fixture(scope="class")
    @register_func(coord_fixtures)
    def simple_dict_coords_dim_name_tuple(self):
        """Create a simple version of coords from dict."""
        input_dict = {
            "time": ("time", self.time),
            "distance": ("distance", self.distance),
        }
        return Coords(input_dict, self.dims)

    @pytest.fixture(scope="class")
    @register_func(coord_fixtures)
    def simple_dict_coords_other_coords(self):
        """Create a simple version of coords from dict."""
        input_dict = {
            "time": ("time", self.time),
            "distance": self.distance,
            "latitude": ("distance", np.ones_like(self.distance)),
        }
        return Coords(input_dict, self.dims)

    @pytest.fixture(scope="class")
    @register_func(coord_fixtures)
    def coord_from_xarray(self, random_patch):
        """Return a Coord instance from xarray coords."""
        dar = random_patch.to_xarray().transpose("time", "distance")
        return Coords(dar.coords)

    @pytest.fixture(scope="class", params=coord_fixtures)
    def coord(self, request):
        """meta-fixture to aggregate all coords."""
        return request.getfixturevalue(request.param)

    def test_dims_inferred(self, coord):
        """Ensure dimensions can be inferred."""
        assert coord.dims == ("time", "distance")

    def test_update(self, coord_from_xarray):
        """Ensure update returns a new coord with updated values."""
        time = coord_from_xarray["time"]
        new_time = time + np.timedelta64(1, "s")
        out = coord_from_xarray.update(time=new_time)
        assert np.all(new_time == out["time"])
        assert isinstance(out, Coords)
        assert out is not coord_from_xarray



"""
Simple tests for FrozenDict.
"""
from collections.abc import Mapping

import pytest

from dascore.utils.mapping import FrozenDict


@pytest.fixture(scope="session")
def frozen_dict():
    """Return an example frozen dict."""
    return FrozenDict({"bob": 1, "bill": 2})


class TestFrozenDict:
    """Test chases for frozen dict"""

    def test_is_mapping(self, frozen_dict):
        """Frozen dict should follow mapping ABC."""
        assert isinstance(frozen_dict, Mapping)

    def test_init_on_dict(self):
        """Ensure a dict can be used to init frozendict."""
        out = FrozenDict({"bob": 1})
        assert isinstance(out, FrozenDict)
        assert "bob" in out

    def test_len(self, frozen_dict):
        """Ensure len works."""
        assert len(frozen_dict) == 2

    def test_contains(self, frozen_dict):
        """Ensure contains works"""
        assert "bob" in frozen_dict
        assert "bill" in frozen_dict

    def test_hash(self, frozen_dict):
        """A frozen dict should be a valid key in a dict/set."""
        out = {frozen_dict: 1}
        assert frozen_dict in out

    def test_init_on_keys(self):
        """Ensure dict can be inited with keys as well."""
        out = FrozenDict(bob=1, bill=2)
        assert isinstance(out, FrozenDict)

    def test_cant_add_keys(self, frozen_dict):
        """Ensure keys can't be added to the dict."""
        with pytest.raises(TypeError, match="not support item assignment"):
            frozen_dict["bob"] = 1

        with pytest.raises(TypeError, match="not support item assignment"):
            frozen_dict["new"] = 1

    def test_cant_mutate_original(self, frozen_dict):
        """
        Ensure the original dict can be changed and this does not affect frozen's
        contents.
        """
        original = {"one": 1, "two": 2}
        froz = FrozenDict(original)
        # test adding new key
        assert "three" not in froz
        original["three"] = 3
        assert "three" not in froz
        # test modifying existing key
        original["one"] = 11
        assert froz["one"] == 1



"""
Implements logic to apply formatting changes to patches from transformations.
"""
import abc
import fnmatch
from abc import abstractmethod


class BaseTransformatter(abc.ABC):
    """
    Base model for helping to apply transformation format changes.
    """

    forward_prefix: str = ""
    inverse_prefix: str = ""

    def _forward_rename(self, name):
        """Rename the dimension for forward transform."""
        if name.startswith(self.inverse_prefix):
            return name[len(self.inverse_prefix) :]
        return f"{self.forward_prefix}{name}"

    def _inverse_rename(self, name):
        """Rename the dimension for backward transform."""
        if name.startswith(self.forward_prefix):
            return name[len(self.forward_prefix) :]
        return f"{self.inverse_prefix}{name}"

    @abstractmethod
    def _forward_unit_label(self, unit_name):
        """Adjust the unit label for forward transform."""

    @abstractmethod
    def _inverse_unit_label(self, unit_name):
        """Adjust the init label for inverse transform."""

    def rename_dims(self, dims, index=None, forward=True):
        """Rename the dimensions."""
        func = self._forward_rename if forward else self._inverse_rename
        new = list(dims)
        index_list = [index] if index is not None else range(len(dims))
        for index in index_list:
            new[index] = func(new[index])
        return tuple(new)

    def rename_attrs(self, dims, attrs, index=None, forward=True):
        """Rename the unit attribute names."""
        func = self._forward_unit_label if forward else self._inverse_unit_label
        index_list = [index] if index is not None else range(len(dims))
        out = dict(attrs)  # this ensures we have a dict and a copy
        for ind in index_list:
            dim_unit_attr_name = f"{dims[ind]}_units"
            if dim_unit_attr_name in out:
                out[dim_unit_attr_name] = func(out[dim_unit_attr_name])
        return out

    def transform_dims_and_attrs(self, dims, attrs, index=None, forward=True):
        """Create new dims and attrs."""
        new_attrs = self.rename_attrs(dims, attrs, index=index, forward=forward)
        new_dims = self.rename_dims(dims, index=index, forward=forward)
        return new_dims, new_attrs


class FourierTransformatter(BaseTransformatter):
    """
    Formatters
    """

    forward_prefix: str = "ft_"
    inverse_prefix: str = "ift_"

    def _toggle_unit_str(self, unit_str):
        """Toggle the unit string."""
        if fnmatch.fnmatch(unit_str, "1/(*)"):
            out = unit_str[3:-1]
        else:
            out = f"1/({unit_str})"
        return out

    def _forward_unit_label(self, unit_name):
        """Adjust the unit label for forward transform."""
        return self._toggle_unit_str(unit_name)

    def _inverse_unit_label(self, unit_name):
        """Adjust the init label for inverse transform."""
        return self._toggle_unit_str(unit_name)



"""
Tests for the spectrogram transformation.
"""
import pytest

import dascore
from dascore.transform.spectro import spectrogram


class TestSpectroTransform:
    """Tests for transforming regular patches into spectrograms."""

    @pytest.fixture()
    def spec_patch(self, random_patch):
        """Simple patch trasnformed to spectrogram."""
        return random_patch.tran.spectrogram("time")

    def test_spec_patch_dimensions(self, spec_patch, random_patch):
        """Ensure expected dimensions now exist."""
        dims = spec_patch.dims
        # dims should have been added
        assert len(dims) > len(random_patch.dims)
        assert set(dims) == (set(random_patch.dims) | {"ft_time"})
        # because of how spectrogram works, the original dim is replaced with
        # the frequency equivalent, and the original dim is added to the end.
        time_index = random_patch.dims.index("time")
        assert spec_patch.dims[time_index] == "ft_time"

    def test_time_first(self, random_patch):
        """Ensure the spectrogram still works when time dim is first."""
        transposed = random_patch.transpose(*("time", "distance"))
        out = spectrogram(transposed, dim="time")
        assert isinstance(out, dascore.Patch)



"""
Tests for reading terra15 format, version 4.
"""
import numpy as np
import pytest

import dascore
from dascore.constants import REQUIRED_DAS_ATTRS
from dascore.core.schema import PatchFileSummary
from dascore.io.terra15.core import Terra15FormatterV4


class TestReadTerra15:
    """Tests for reading the terra15 format."""

    def test_type(self, terra15_das_patch):
        """Ensure the expected type is returned."""
        assert isinstance(terra15_das_patch, dascore.Patch)

    def test_attributes(self, terra15_das_patch):
        """Ensure a few of the expected attrs exist in array."""
        attrs = dict(terra15_das_patch.attrs)
        expected_attrs = {"time_min", "time_max", "distance_min", "data_units"}
        assert set(expected_attrs).issubset(set(attrs))

    def test_has_required_attrs(self, terra15_das_patch):
        """Ensure the required das attrs are found."""
        assert set(REQUIRED_DAS_ATTRS).issubset(set(dict(terra15_das_patch.attrs)))

    def test_coord_attr_time_equal(self, terra15_das_patch):
        """The time reported in the attrs and coords should match."""
        attr_time = terra15_das_patch.attrs["time_max"]
        coord_time = terra15_das_patch.coords["time"].max()
        assert attr_time == coord_time

    def test_time_dist_slice(self, terra15_das_patch, terra15_das_example_path):
        """Ensure slicing distance and time works from read func."""
        time_array = terra15_das_patch.coords["time"]
        dist_array = terra15_das_patch.coords["distance"]
        t1, t2 = time_array[10], time_array[40]
        d1, d2 = dist_array[10], dist_array[40]
        patch = Terra15FormatterV4().read(
            terra15_das_example_path, time=(t1, t2), distance=(d1, d2)
        )[0]
        attrs, coords = patch.attrs, patch.coords
        assert attrs["time_min"] == coords["time"].min() == t1
        assert (attrs["time_max"] - t2) < (attrs["d_time"] / 4)
        assert attrs["distance_min"] == coords["distance"].min() == d1
        assert attrs["distance_max"] == coords["distance"].max() == d2

    def test_no_arrays_in_attrs(self, terra15_das_patch):
        """
        Ensure that the attributes are not arrays.
        Originally, attrs like time_min can be arrays with empty shapes.
        """
        for key, val in terra15_das_patch.attrs.items():
            assert not isinstance(val, np.ndarray)


class TestIsTerra15:
    """Tests for function to determine if a file is a terra15 file."""

    def test_format_and_version(self, terra15_das_example_path):
        """Ensure version two is recognized."""
        format, version = Terra15FormatterV4().get_format(terra15_das_example_path)
        assert format == Terra15FormatterV4.name
        assert version == Terra15FormatterV4.version

    def test_not_terra15_not_df5(self, dummy_text_file):
        """Test for not even a hdf5 file."""
        parser = Terra15FormatterV4()
        assert not parser.get_format(dummy_text_file)
        assert not parser.get_format(dummy_text_file.parent)

    def test_hdf5file_not_terra15(self, generic_hdf5):
        """Assert that the generic hdf5 file is not a terra15."""
        parser = Terra15FormatterV4()
        assert not parser.get_format(generic_hdf5)


class TestScanTerra15:
    """Tests for scanning terra15 file."""

    def test_scanning(self, terra15_das_patch, terra15_das_example_path):
        """Tests for getting summary info from terra15 data."""
        parser = Terra15FormatterV4()
        out = parser.scan(terra15_das_example_path)
        assert isinstance(out, list)
        assert len(out) == 1
        assert isinstance(out[0], PatchFileSummary)

        data = out[0]
        assert data.file_format == parser.name
        assert data.file_version == parser.version


class TestTerra15Unfinished:
    """Test for reading files with zeroes filled at the end."""

    @pytest.fixture(scope="class")
    def patch_unfinished(self, terra15_das_unfinished_path):
        """Return the patch with zeroes at the end."""
        parser = Terra15FormatterV4()
        out = parser.read(terra15_das_unfinished_path)[0]
        return out

    def test_zeros_gone(self, patch_unfinished):
        """No zeros should exist in the data."""
        data = patch_unfinished.data
        all_zero_rows = np.all(data == 0, axis=1)
        assert not np.any(all_zero_rows)

    def test_monotonic_time(self, patch_unfinished):
        """Ensure the time is increasing."""
        time = patch_unfinished.coords["time"]
        assert np.all(np.diff(time) >= np.timedelta64(0, "s"))



"""DASCore - A library for fiber optic sensing."""
# set xarray settings
from xarray import set_options

from dascore.core.patch import Patch
from dascore.core.schema import PatchAttrs
from dascore.core.spool import BaseSpool, spool
from dascore.examples import get_example_patch, get_example_spool
from dascore.io.core import get_format, read, scan, scan_to_df, write
from dascore.utils.patch import patch_function
from dascore.utils.time import to_datetime64, to_timedelta64
from dascore.version import __last_version__, __version__

# keep attrs on xarray DataArray
set_options(keep_attrs=True)

# flag for disabling progress bar when debugging
_debug = False



"""
Tests for converting velocity to strain-rate.
"""
import numpy as np
import pytest

from dascore.exceptions import PatchAttributeError


class TestStrainRateConversion:
    """Tests for converting velocity to strain-rate."""

    @pytest.fixture()
    def patch_strain_rate_default(self, terra15_das_patch):
        """Return the default terra15 converted to strain rate."""
        return terra15_das_patch.tran.velocity_to_strain_rate()

    def test_attrs(self, patch_strain_rate_default):
        """Ensure the attributes were updated with strain_rate"""
        attrs = patch_strain_rate_default.attrs
        assert attrs["data_type"] == "strain_rate"

    def test_data_different(self, patch_strain_rate_default, terra15_das_patch):
        """The data should have changed."""
        data1 = patch_strain_rate_default.data
        data2 = terra15_das_patch.data
        assert not np.all(np.equal(data1, data2))

    def test_raises_on_strain_rate(self, patch_strain_rate_default):
        """It does not make sense to apply this twice."""
        with pytest.raises(PatchAttributeError, match="velocity"):
            _ = patch_strain_rate_default.tran.velocity_to_strain_rate()



"""
Module for spools, containers of patches.
"""
import abc
from functools import singledispatch
from pathlib import Path
from typing import Mapping, Optional, Sequence, Union

import pandas as pd
from typing_extensions import Self

import dascore as dc
from dascore.constants import PatchType, SpoolType, numeric_types, timeable_types
from dascore.exceptions import InvalidSpoolError
from dascore.utils.chunk import ChunkManager
from dascore.utils.docs import compose_docstring
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import CacheDescriptor
from dascore.utils.patch import _merge_patches, patches_to_df
from dascore.utils.pd import (
    _convert_min_max_in_kwargs,
    adjust_segments,
    filter_df,
    get_column_names_from_dim,
    get_dim_names_from_columns,
)


class BaseSpool(abc.ABC):
    """
    Spool Abstract Base Class (ABC) for defining Spool interface.
    """

    @abc.abstractmethod
    def __getitem__(self, item: int) -> PatchType:
        """Returns a patch from the spool."""

    @abc.abstractmethod
    def __iter__(self) -> PatchType:
        """Iterate through the Patches in the spool."""

    def update(self: SpoolType) -> Self:
        """
        Updates the contents of the spool and returns a spool.
        """
        return self

    @abc.abstractmethod
    def chunk(
        self: SpoolType,
        overlap: Optional[Union[numeric_types, timeable_types]] = None,
        keep_partial: bool = False,
        snap_coords: bool = True,
        tolerance: float = 1.5,
        **kwargs,
    ) -> Self:
        """
        Chunk the data in the spool along specified dimensions.

        Parameters
        ----------
        overlap
            The amount of overlap between each segment, starting with the end of
            first patch. Negative values can be used to induce gaps.
        keep_partial
            If True, keep the segments which are smaller than chunk size.
            This often occurs because of data gaps or at end of chunks.
        snap_coords
            If True, snap the coords on joined patches such that the spacing
            remains constant.
        tolerance
            The number of samples a block of data can be spaced and still be
            considered contiguous.
        kwargs
            kwargs are used to specify the dimension along which to chunk, eg:
            `time=10` chunks along the time axis in 10 second increments.
        """

    @abc.abstractmethod
    def select(self: SpoolType, **kwargs) -> Self:
        """
        Select only part of the data.
        """

    @abc.abstractmethod
    def get_contents(self: SpoolType) -> pd.DataFrame:
        """
        Get a dataframe of the patches that will be returned by the spool.
        """

    def __len__(self):
        pass


class DataFrameSpool(BaseSpool):
    """
    An abstract class for spools whose contents are managed by a dataframe.
    """

    # A dataframe which represents contents as they will be output
    _df: pd.DataFrame = CacheDescriptor("_cache", "_get_df")
    # A dataframe which shows patches in the source
    _source_df: pd.DataFrame = CacheDescriptor("_cache", "_get_source_df")
    # A dataframe of instructions for going from source_df to df
    _instruction_df: pd.DataFrame = CacheDescriptor("_cache", "_get_instruction_df")
    # kwargs for filtering contents
    _select_kwargs: Optional[Mapping] = FrozenDict()
    # attributes which effect merge groups for internal patches
    _group_columns = ("network", "station", "dims", "data_type", "history", "tag")
    _drop_columns = ("patch",)

    def _get_df(self):
        """Function to get the current df."""

    def _get_source_df(self):
        """Function to get the current df."""

    def _get_instruction_df(self):
        """Function to get the current df."""

    def __init__(self):
        self._cache = {}

    def __getitem__(self, item):
        out = self._get_patches_from_index(item)
        # a single index was used, should return a single patch
        if not isinstance(item, slice):
            out = self._unbox_patch(out)
        # a slice was used, return a sub-spool
        else:
            out = self.__class__(out)
        return out

    def __len__(self):
        return len(self._df)

    def __iter__(self):
        for ind in self._df.index:
            yield self._unbox_patch(self._get_patches_from_index(ind))

    def _unbox_patch(self, patch_list):
        """Unbox a single patch from a patch list, check len."""
        assert len(patch_list) == 1
        return patch_list[0]

    def _get_patches_from_index(self, df_ind):
        """
        Given an index (from current df), return the corresponding patch.
        """
        source = self._source_df
        instruction = self._instruction_df
        if isinstance(df_ind, slice):  # handle slicing
            df1 = instruction.loc[instruction["current_index"].values[df_ind]]
        else:  # Filter instruction df to only include current index.
            df1 = instruction[instruction["current_index"] == df_ind]
        if df1.empty:
            msg = f"index of [{df_ind}] is out of bounds for spool."
            raise IndexError(msg)
        joined = df1.join(source.drop(columns=df1.columns, errors="ignore"))
        return self._patch_from_instruction_df(joined)

    def _patch_from_instruction_df(self, joined):
        """Get the patches joined columns of instruction df."""
        df_dict_list = self._df_to_dict_list(joined)
        out_list = []
        expected_len = len(joined["current_index"].unique())
        for patch_kwargs in df_dict_list:
            # convert kwargs to format understood by parser/patch.select
            kwargs = _convert_min_max_in_kwargs(patch_kwargs, joined)
            patch = self._load_patch(kwargs)
            select_kwargs = {
                i: v for i, v in kwargs.items() if i in patch.dims or i in patch.coords
            }
            out_list.append(patch.select(**select_kwargs))
        if len(out_list) > expected_len:
            patch_df = patches_to_df(out_list)
            # Tolerance should be high since the dataframe chunker has
            # already determined these rows should be merged together.
            out_list = _merge_patches(patch_df, tolerance=1_000_000)
        return out_list

    @staticmethod
    def _get_dummy_dataframes(input_df):
        """
        Return dummy current, source, and instruction dataframes.

        Dummy because the source and current df are the same, so the
        instruction df is a straight mapping between the two.
        """
        output = input_df.copy(deep=False)
        dims = get_dim_names_from_columns(output)
        cols2keep = get_column_names_from_dim(dims)
        instruction = (
            input_df.copy()[cols2keep]
            .assign(source_index=output.index, current_index=output.index)
            .set_index("source_index")
            .sort_values("current_index")
        )
        return input_df, output, instruction

    def _df_to_dict_list(self, df):
        """
        Convert the dataframe to a list of dicts for iteration.

        This is significantly faster than iterating rows.
        """
        df_dict_list = list(df.T.to_dict().values())
        return df_dict_list

    @abc.abstractmethod
    def _load_patch(self, kwargs) -> Self:
        """Given a row from the managed dataframe, return a patch."""

    @compose_docstring(doc=BaseSpool.chunk.__doc__)
    def chunk(
        self: SpoolType,
        overlap: Optional[Union[numeric_types, timeable_types]] = None,
        keep_partial: bool = False,
        snap_coords: bool = True,
        tolerance: float = 1.5,
        **kwargs,
    ) -> Self:
        """
        {doc}
        """
        df = self._df.drop(columns=list(self._drop_columns), errors="ignore")
        chunker = ChunkManager(
            overlap=overlap,
            keep_partial=keep_partial,
            group_columns=self._group_columns,
            tolerance=tolerance,
            **kwargs,
        )
        in_df, out_df = chunker.chunk(df)
        if df.empty:
            instructions = None
        else:
            instructions = chunker.get_instruction_df(in_df, out_df)
        return self.new_from_df(out_df, source_df=self._df, instruction_df=instructions)

    def new_from_df(self, df, source_df=None, instruction_df=None):
        """Create a new instance from dataframes."""
        new = self.__class__(self)
        df_, source_, inst_ = self._get_dummy_dataframes(df)
        new._df = df
        new._source_df = source_df if source_df is not None else source_
        new._instruction_df = instruction_df if instruction_df is not None else inst_
        return new

    def select(self, **kwargs) -> Self:
        """Sub-select certain dimensions for Spool"""
        inst = adjust_segments(self._instruction_df, ignore_bad_kwargs=True, **kwargs)
        out = self.new_from_df(
            adjust_segments(self._df, **kwargs),
            source_df=self._source_df,
            instruction_df=inst,
        )
        return out

    @compose_docstring(doc=BaseSpool.get_contents.__doc__)
    def get_contents(self) -> pd.DataFrame:
        """
        {doc}
        """
        return self._df[filter_df(self._df, **self._select_kwargs)]


class MemorySpool(DataFrameSpool):
    """
    A Spool for storing patches in memory.
    """

    # tuple of attributes to remove from table

    def __init__(self, data: Optional[Union[PatchType, Sequence[PatchType]]] = None):
        super().__init__()
        if data is not None:
            dfs = self._get_dummy_dataframes(patches_to_df(data))
            self._df, self._source_df, self._instruction_df = dfs

    def __str__(self):
        """Returns a (hopefully) useful string rep of spool."""
        df = self._df
        tmin, tmax = df["time_min"].min(), df["time_max"].max()
        out = (
            f"MemorySpool object managing {len(self)} patches spanning:"
            f" {tmin} to {tmax}"
        )
        return out

    __repr__ = __str__

    def _load_patch(self, kwargs) -> Self:
        """Load the patch into memory"""
        return kwargs["patch"]


@singledispatch
def spool(obj: Union[Path, str, BaseSpool, Sequence[PatchType]], **kwargs) -> SpoolType:
    """
    Load a spool from some data source.

    Parameters
    ----------
    obj
        An object from which a spool can be derived.
    """
    msg = f"Could not get spool from: {obj}"
    raise ValueError(msg)


@spool.register(str)
@spool.register(Path)
def spool_from_str(path, **kwargs):
    """Get a spool from a path."""
    path = Path(path)
    # A directory was passed, create Directory Spool
    if path.is_dir():
        from dascore.clients.dirspool import DirectorySpool

        return DirectorySpool(path, **kwargs)
    # A single file was passed. If the file format supports quick scanning
    # Return a FileSpool (lazy file reader), else return DirectorySpool.
    elif path.exists():  # a single file path was passed.
        from dascore.io.core import FiberIO

        _format, _version = dc.get_format(path)
        formatter = FiberIO.manager.get_fiberio(_format, _version)
        if formatter.implements_scan:
            from dascore.clients.filespool import FileSpool

            return FileSpool(path, _format, _version)

        else:
            return MemorySpool(dc.read(path, _format, _version))
    else:
        msg = (
            f"could not get spool from argument: {path}. "
            f"If it is a path, it may not exist."
        )
        raise InvalidSpoolError(msg)


@spool.register(BaseSpool)
def spool_from_spool(spool, **kwargs):
    """Return a spool from a spool."""
    return spool


@spool.register(list)
@spool.register(tuple)
def spool_from_patch_list(patch_list, **kwargs):
    """Return a spool from a sequence of patches."""
    return MemorySpool(patch_list)


@spool.register(dc.Patch)
def spool_from_patch(patch):
    """Get a spool from a single patch."""
    return MemorySpool([patch])



"""
Module to Patch Processing.
"""
from .aggregate import aggregate
from .basic import abs, normalize, rename, squeeze, standardize, transpose
from .detrend import detrend
from .filter import median_filter, pass_filter, sobel_filter
from .resample import decimate, interpolate, iresample, resample
from .select import select



"""
Modules for Fourier transforms.
"""
import numpy as np

from dascore.constants import PatchType
from dascore.utils.misc import _get_sampling_rate
from dascore.utils.patch import patch_function
from dascore.utils.transformatter import FourierTransformatter


@patch_function()
def rfft(patch: PatchType, dim="time") -> PatchType:
    """
    Perform a real fourier transform along the specified dimension.

    Parameters
    ----------
    patch
        The Patch object to transform.
    dim
        The dimension along which the transform is applied.

    Notes
    -----
    The real Fourier Transform is only appropriate for real-valued data arrays.
    """
    ft = FourierTransformatter()
    data = patch.data
    axis = patch.dims.index(dim)
    sr = 1 / _get_sampling_rate(patch.attrs[f"d_{dim}"])
    freqs = np.fft.rfftfreq(data.shape[axis], sr)
    new_data = np.fft.rfft(data, axis=axis)
    new_dim_name = ft._forward_rename(patch.dims[axis])
    dims, attrs = ft.transform_dims_and_attrs(patch.dims, patch.attrs, index=axis)
    new_coords = {new_dim_name: freqs}
    new_coords.update({x: patch.coords[x] for x in patch.dims if x != dim})
    return patch.__class__(data=new_data, coords=new_coords, dims=dims, attrs=attrs)



"""
pytest configuration for dascore
"""
import os
import shutil
from pathlib import Path

import matplotlib
import numpy as np
import pytest
import tables as tb
import tables.parameters

import dascore as dc
import dascore.examples as ex
from dascore.clients.dirspool import DirectorySpool
from dascore.constants import SpoolType
from dascore.core import Patch
from dascore.io.core import read
from dascore.utils.downloader import fetch
from dascore.utils.misc import register_func

test_data_path = Path(__file__).parent.absolute() / "test_data"

# A list to register functions that return general spools or patches
# These are to be used for running many patches/spools through
# Generic tests.
SPOOL_FIXTURES = []
PATCH_FIXTURES = []


# --- Pytest configuration


def pytest_addoption(parser):
    """Add pytest command options."""
    parser.addoption(
        "--integration",
        action="store_true",
        dest="run_integration",
        default=False,
        help="Run integration tests",
    )


def pytest_collection_modifyitems(config, items):
    """Configure pytest command line options."""

    marks = {}
    if not config.getoption("--integration"):
        msg = "needs --integration option to run"
        marks["integration"] = pytest.mark.skip(reason=msg)

    for item in items:
        marks_to_apply = set(marks)
        item_marks = set(item.keywords)
        for mark_name in marks_to_apply & item_marks:
            item.add_marker(marks[mark_name])


def pytest_sessionstart(session):
    """
    Hook to run before any other tests.

    Used to ensure a non-visual backend is used so plots don't pop up
    and to set debug hook to True to avoid showing progress bars,
    except when explicitly being tested.
    """
    # If running in CI make sure to turn off matplotlib.
    if os.environ.get("CI", False):
        matplotlib.use("Agg")

    # need to set nodes to 32 to avoid crash on p3.11. See pytables#977.
    tables.parameters.NODE_CACHE_SLOTS = 32

    # Ensure debug is set. This disables progress bars which disrupt debugging.
    dc._debug = True


# --- Patch Paths


@pytest.fixture(scope="session")
def terra15_das_example_path():
    """Return the path to the example terra15 file."""
    out = fetch("terra15_das_1_trimmed.hdf5")
    assert out.exists()
    return out


@pytest.fixture(scope="session")
@register_func(SPOOL_FIXTURES)
def terra15_das_unfinished_path() -> Path:
    """Return the spool of Terra15 Das Array"""
    out = fetch("terra15_das_unfinished.hdf5")
    assert out.exists()
    return out


@pytest.fixture(scope="class")
def terra15_v5_path():
    """Get the path to terra15 V5 file, download if not cached."""
    return fetch("terra15_v5_test_file.hdf5")


@pytest.fixture(scope="class")
def terra15_v6_path():
    """Get the path to terra15 V5 file, download if not cached."""
    return fetch("terra15_v6_test_file.hdf5")


@pytest.fixture()
@register_func(SPOOL_FIXTURES)
def terra15_das_spool(terra15_das_example_path) -> SpoolType:
    """Return the spool of Terra15 Das Array"""
    return read(terra15_das_example_path, file_format="terra15")


# --- Patch fixtures


@pytest.fixture(scope="session")
@register_func(PATCH_FIXTURES)
def terra15_das_patch(terra15_das_example_path) -> Patch:
    """Read the terra15 data, return contained DataArray"""
    out = read(terra15_das_example_path, "terra15")[0]
    attr_time = out.attrs["time_max"]
    coord_time = out.coords["time"].max()
    assert attr_time == coord_time
    return out


@pytest.fixture(scope="session")
@register_func(PATCH_FIXTURES)
def random_patch() -> Patch:
    """Init a random array."""
    from dascore.examples import get_example_patch

    return get_example_patch("random_das")


@pytest.fixture(scope="class")
@register_func(PATCH_FIXTURES)
def random_patch_with_lat_lon(random_patch):
    """Get a random patch with added lat/lon coordinates."""
    return dc.get_example_patch("random_patch_with_lat_lon")


@pytest.fixture(scope="class")
@register_func(PATCH_FIXTURES)
def multi_dim_coords_patch(random_patch):
    """A patch with a multiple dimensional coord."""
    dist = random_patch.coords["distance"]
    time = random_patch.coords["time"]
    quality = np.ones((len(dist), len(time)))
    out = random_patch.assign_coords(quality=(("distance", "time"), quality))
    return out


@pytest.fixture(scope="session")
@register_func(PATCH_FIXTURES)
def event_patch_1():
    """Fetch event patch 1."""
    path = fetch("example_dasdae_event_1.h5")
    return dc.spool(path)[0]


@pytest.fixture(scope="class", params=PATCH_FIXTURES)
def patch(request):
    """A meta-fixtures for collecting all patches used in testing."""
    return request.getfixturevalue(request.param)


# --- Spool setup fixtures


@pytest.fixture(scope="class")
def one_file_dir(tmp_path_factory, random_patch):
    """Create a directory with a single DAS file."""
    out = Path(tmp_path_factory.mktemp("one_file_file_spool"))
    spool = dc.spool(random_patch)
    return ex.spool_to_directory(spool, path=out)


@pytest.fixture(scope="class")
def two_patch_directory(tmp_path_factory, terra15_das_example_path, random_patch):
    """Create a directory of DAS files for testing."""
    # first copy in a terra15 file
    dir_path = tmp_path_factory.mktemp("bank_basic")
    shutil.copy(terra15_das_example_path, dir_path)
    # save a random patch
    random_patch.io.write(dir_path / "random.hdf5", "dasdae")
    return dir_path


@pytest.fixture(scope="class")
def diverse_spool_directory(diverse_spool):
    """Save the diverse spool contents to a directory."""
    out = ex.spool_to_directory(diverse_spool)
    yield out
    if out.is_dir():
        shutil.rmtree(out)


@pytest.fixture(scope="class")
def adjacent_spool_directory(tmp_path_factory, adjacent_spool_no_overlap):
    """Create a directory of adjacent patches."""
    # create a directory with several patch files in it.
    dir_path = Path(tmp_path_factory.mktemp("data"))
    for num, patch in enumerate(adjacent_spool_no_overlap):
        path = dir_path / f"{num}_patch.hdf5"
        dc.write(patch, path, file_format="dasdae")
    return dir_path


# --- Spool fixtures


@pytest.fixture(scope="session")
@register_func(SPOOL_FIXTURES)
def random_spool() -> SpoolType:
    """Init a random array."""
    from dascore.examples import get_example_spool

    return get_example_spool("random_das")


@pytest.fixture(scope="class")
@register_func(SPOOL_FIXTURES)
def adjacent_spool_no_overlap(random_patch) -> dc.BaseSpool:
    """
    Create a spool with several patches within one time sample but not
    overlapping.
    """
    pa1 = random_patch
    t2 = random_patch.attrs["time_max"]
    d_time = random_patch.attrs["d_time"]

    pa2 = random_patch.update_attrs(time_min=t2 + d_time)
    t3 = pa2.attrs["time_max"]

    pa3 = pa2.update_attrs(time_min=t3 + d_time)

    expected_time = pa3.attrs["time_max"] - pa1.attrs["time_min"]
    actual_time = pa3.coords["time"].max() - pa1.coords["time"].min()
    assert expected_time == actual_time
    return dc.spool([pa2, pa1, pa3])


@pytest.fixture(scope="class")
@register_func(SPOOL_FIXTURES)
def one_file_directory_spool(one_file_dir):
    """Create a directory with a single DAS file."""
    return DirectorySpool(one_file_dir).update()


@pytest.fixture(scope="class")
@register_func(SPOOL_FIXTURES)
def diverse_spool():
    """Create a spool with a diverse set of patches for testing."""
    return ex._diverse_spool()


@pytest.fixture(scope="class")
@register_func(SPOOL_FIXTURES)
def diverse_directory_spool(diverse_spool_directory):
    """Save the diverse spool contents to a directory."""
    out = dc.spool(diverse_spool_directory).update()
    return out


@pytest.fixture(scope="class")
@register_func(SPOOL_FIXTURES)
def basic_file_spool(two_patch_directory):
    """Return a DAS bank on basic_bank_directory."""
    out = DirectorySpool(two_patch_directory)
    return out.update()


@pytest.fixture(scope="class", params=SPOOL_FIXTURES)
def spool(request):
    """A meta-fixtures for collecting all spools used in testing."""
    return request.getfixturevalue(request.param)


# --- Misc. test fixtures


@pytest.fixture
def generic_hdf5(tmp_path):
    """
    Create a generic hdf5 file (not das). This is useful for ensuring formatters
    recognize differences in HDF5 files.
    """
    parent = tmp_path / "sum"
    parent.mkdir()
    path = parent / "simple.hdf5"

    with tb.open_file(str(path), "w") as fi:
        group = fi.create_group("/", "bob")
        fi.create_carray(group, "data", obj=np.random.rand(10))
    return path


@pytest.fixture(scope="class")
def dummy_text_file(tmp_path_factory):
    """Return a text file with silliness in it."""
    parent = tmp_path_factory.mktemp("dummy")
    path = parent / "hello.txt"
    path.write_text("Clearly not a hdf5 file. Or is it?")
    return path



"""
Utilities for working with coordinates on Patches.
"""
from __future__ import annotations

from typing_extensions import Self

from dascore.constants import PatchType
from dascore.utils.mapping import FrozenDict


def assign_coords(patch: PatchType, **kwargs) -> PatchType:
    """
    Add non-dimensional coordinates to a patch.

    Parameters
    ----------
    patch
        The patch to which coordinates will be added.
    **kwargs
        Used to specify the name, dimension, and values of the new
        coordinates.

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> patch_1 = dc.get_example_patch()
    >>> coords = patch_1.coords
    >>> dist = coords['distance']
    >>> time = coords['time']
    >>> # Add a single coordinate associated with distance dimension
    >>> lat = np.arange(0, len(dist)) * .001 -109.857952
    >>> out_1 = patch_1.assign_coords(latitude=('distance', lat))
    >>> # Add multiple coordinates associated with distance dimension
    >>> lon = np.arange(0, len(dist)) *.001 + 41.544654
    >>> out_2 = patch_1.assign_coords(
    ...     latitude=('distance', lat),
    ...     longitude=('distance', lon),
    ... )
    >>> # Add multi-dimensional coordinates
    >>> quality = np.ones_like(patch_1.data)
    >>> out_3 = patch_1.assign_coords(
    ...     quality=(patch_1.dims, quality)
    ... )
    """
    coords = {x: patch.coords[x] for x in patch.coords}
    for coord_key, (dimension, value) in kwargs.items():
        coords[coord_key] = (dimension, value)
    return patch.new(coords=coords, dims=patch.dims)


class Coords:
    """
    A class to simplify the handling of coordinates.

    Also helps in supporting non-dimensional coordinates and inferring
    dimensions.

    Attributes
    ----------
    array_dict
        A dict of {coord_name: array}.
    dims_dict
        A dict of {coord_name: tuple[dim_1, dim_2...]}.
    dims
        A tuple of the dimensions for the coordinates.
    """

    # --- Init stuff
    dims = ()
    array_dict = FrozenDict()
    dims_dict = FrozenDict()

    def __init__(self, coords, dims=None):
        if coords is None:
            return
        # Another coord as input
        if isinstance(coords, Coords):
            self.__dict__.update(coords.__dict__)
            return
        # hande a dict being passed
        if isinstance(coords, (dict, FrozenDict)):
            array_dict, dim_dict = self._coord_dict_from_dict(coords)
            self.dims = dims
        # handle xarray coordinate
        else:
            array_dict, dim_dict = self._coord_dict_from_xr(coords)
            self.dims = coords.dims
        self.array_dict = FrozenDict(array_dict)
        self.dims_dict = FrozenDict(dim_dict)

    def _coord_dict_from_xr(self, coord):
        """Get the coordinates from an xarray coordinate object."""
        array_dict = {}
        dim_dict = {}
        for i, v in coord.items():
            array_dict[i] = v.data
            dims = v.dims
            dim_dict[i] = (dims,) if isinstance(dims, str) else tuple(dims)
        return array_dict, dim_dict

    def _coord_dict_from_dict(self, coord_dict):
        """Get the coordinate dict from a dictionary"""
        arrays = {}
        dimensions = {}
        for i, v in coord_dict.items():
            # determine if of the form (dim_name(s), value) or just value
            if len(v) and isinstance(v[0], (str, tuple, list)):
                assert len(v) == 2  # should be a sequence of dims, array
                arrays[i] = v[1]
                dims = v[0]
                dimensions[i] = (dims,) if isinstance(dims, str) else tuple(dims)
            else:
                arrays[i] = v
                dimensions[i] = (i,)
        return arrays, dimensions

    def __str__(self):
        # TODO: Once I am convinced Coords will stay we need a better str
        msg = f"Coords managing {list(self.dims_dict)}"
        return msg

    __repr__ = __str__

    def get(self, item, default=None):
        """Return item or None if not in coord. Same as dict.get."""
        return self.array_dict.get(item, default=default)

    def __iter__(self):
        return self.array_dict.__iter__()

    def __getitem__(self, item):
        return self.array_dict[item]

    def to_nested_dict(self):
        """
        Return a dict of {coord_name: ((*dims), data)}.

        This is useful for input into xarray.
        """
        out = {i: (self.dims_dict[i], self.array_dict[i]) for i in self.dims_dict}
        return out

    def update(self, **kwargs) -> Self:
        """Set the data of an existing coordinate."""
        assert set(kwargs).issubset(set(self.array_dict))
        coords = self.to_nested_dict()
        for item, val in kwargs.items():
            data_list = list(coords[item])
            data_list[1] = val
            coords[item] = data_list
        out = self.__class__(coords)
        return out



"""
Pandas utilities.
"""
import fnmatch
import os
from collections import defaultdict
from functools import cache
from typing import Collection, Sequence, Tuple, Type

import numpy as np
import pandas as pd
from pydantic import BaseModel

from dascore.utils.time import to_datetime64, to_timedelta64


@cache
def get_regex(seed_str):
    """Compile, and cache regex for str queries."""
    return fnmatch.translate(seed_str)  # translate to re


def _remove_base_path(series: pd.Series, base="") -> pd.Series:
    """
    Ensure paths stored in column name use unix style paths and have base
    path removed.
    """
    if series.empty:
        return series
    unix_paths = series.str.replace(os.sep, "/")
    unix_base_path = str(base).replace(os.sep, "/")
    return unix_paths.str.replace(unix_base_path, "", regex=False)


def _get_min_max_query(kwargs, df):
    """
    Get a dict of {column_name: Optional[min_val], Optional[max_val]}.

    Handles {column}_max, column_{min} type queries. Pop keys out of kwargs
    once they are in the return dict.
    """
    out = defaultdict(lambda: [None, None])
    col_set = set(df.columns)
    to_kill = []
    for key, val in kwargs.items():
        if key.endswith("_max") and key not in col_set:
            out[key.replace("_max", "")][1] = val
            to_kill.append(key)
        elif key.endswith("_min") and key not in col_set:
            out[key.replace("_min", "")][0] = val
            to_kill.append(key)
    # remove keys with min/max suffix
    for key in to_kill:
        kwargs.pop(key, None)
    return out


def _add_range_query(kwargs, df, ignore_bad_kwargs=False):
    """
    Add a range query that spans two columns.

    For example, if columns 'time_min' and 'time_max' exist but 'time'
    does not, time=(time_1, time_2) will filter df to only include columns
    which have a range in specified time.
    """
    col_set = set(df)
    unknown_cols = set(kwargs) - col_set
    bad_keys = set()
    range_query = {}
    for key in unknown_cols:
        min_key, max_key = f"{key}_min", f"{key}_max"
        val = kwargs[key]
        if {min_key, max_key}.issubset(col_set) and len(val) == 2:
            range_query[key] = val
            kwargs.pop(key, None)
        else:
            bad_keys.add(key)
    if len(bad_keys):
        if not ignore_bad_kwargs:
            msg = f"columns: {bad_keys} are not found in df"
            raise KeyError(msg)
        else:
            for key in bad_keys:
                kwargs.pop(key, None)
    return range_query


def _get_flat_and_collection_queries(kwargs):
    """Divide kwargs into flat and sequence queries."""
    flat_query = {
        k: v
        for k, v in kwargs.items()
        if isinstance(v, str) or not isinstance(v, Collection)
    }
    sequence_query = {
        k: v for k, v in kwargs.items() if k not in flat_query and v is not None
    }
    return flat_query, sequence_query


def _filter_equality(query_dict, df, bool_index):
    """Filter based on equality checks."""
    # filter on non-collection queries
    for key, val in query_dict.items():
        if isinstance(val, str):
            regex = get_regex(val)
            new = df[key].str.match(regex).values
            bool_index = np.logical_and(bool_index, new)
        else:
            new = (df[key] == val).values
            bool_index = np.logical_and(bool_index, new)
    return bool_index


def _filter_contains(query_dict, df, bool_index):
    """Filter based on rows containing specified values."""
    for key, val in query_dict.items():
        bool_index = np.logical_and(bool_index, df[key].isin(val))
    return bool_index


def _filter_range(query_dict, df, bool_index):
    """Filter based on ranges for columns."""
    for key, (min_val, max_val) in query_dict.items():
        col = df[key]
        if min_val is not None:
            bool_index = np.logical_and(bool_index, col >= min_val)
        if max_val is not None:
            bool_index = np.logical_and(bool_index, col <= max_val)
    return bool_index


def _filter_multicolumn_range(query_dict, df, bool_index):
    """Filter based on inclusive ranges in multiple columns."""
    for key, val in query_dict.items():
        min_key, max_key = f"{key}_min", f"{key}_max"
        min_col, max_col = df[min_key], df[max_key]
        if val[0] is not None:
            max_too_big = max_col < val[0]
            bool_index = np.logical_and(~max_too_big, bool_index)
        if val[1] is not None:
            min_too_small = min_col > val[1]
            bool_index = np.logical_and(~min_too_small, bool_index)
    return bool_index


def _convert_times(df, some_dict):
    """Convert query values to datetime/timedelta values."""
    if not some_dict:
        return some_dict
    # convert queries related to datetime into datetime64
    datetime_cols = set(df.select_dtypes(include=np.datetime64).columns)
    non_min_max_cols = {x.replace("_min", "") for x in datetime_cols}
    datetime_keys = (datetime_cols & set(some_dict)) | (
        non_min_max_cols & set(some_dict)
    )
    for key in datetime_keys:
        some_dict[key] = to_datetime64(some_dict[key])
    # convert queries related to time delta into timedelta64
    timedelta_cols = set(df.select_dtypes(include=np.timedelta64).columns)
    timedelta_keys = timedelta_cols & set(some_dict)
    for key in timedelta_keys:
        some_dict[key] = to_timedelta64(some_dict[key])
    return some_dict


def get_interval_columns(df, name):
    """Return a series of start, stop, step for columns."""
    names = f"{name}_min", f"{name}_max", f"d_{name}"
    missing_cols = set(names) - set(df.columns)
    if missing_cols:
        msg = f"Dataframe is missing {missing_cols} to chunk on {name}"
        raise KeyError(msg)
    return df[names[0]], df[names[1]], df[names[2]]


def yield_slice_from_kwargs(df, kwargs) -> Tuple[str, slice]:
    """
    For each slice keyword, yield the name and slice.

    Will also convert slice values based on dtypes in dataframe, eg
    time=(1, 10) will convert to
    time=(np.timedelta64(1, 's'), np.timedelta64(10, 's')) provided columns
    'time_min' and 'time_max' are datetime columns.
    """

    def _get_slice(value):
        """Ensure the value can rep. a slice."""
        assert isinstance(value, (slice, Sequence)) and len(value) == 2
        if not isinstance(value, slice):
            value = slice(*value)
        return value

    def _maybe_convert_dtype(sli, name, df):
        """Convert dtypes of slice if needed."""
        datetime_cols = set(df.select_dtypes(include=np.datetime64).columns)
        if {f"{name}_min", f"{name}_max"}.issubset(datetime_cols):
            sli = slice(
                to_datetime64(sli.start) if sli.start is not None else None,
                to_datetime64(sli.stop) if sli.stop is not None else None,
                to_timedelta64(sli.step) if sli.step is not None else None,
            )
        return sli

    # find keys which correspond to column ranges
    col_set = set(df.columns)
    valid_minmax_kwargs = {
        x
        for x in kwargs
        if {f"{x}_min", f"{x}_max"}.issubset(col_set) and x not in col_set
    }
    # ensure exactly one column is found
    for name in valid_minmax_kwargs:
        out_slice = _maybe_convert_dtype(_get_slice(kwargs[name]), name, df)
        yield name, out_slice


def adjust_segments(df, ignore_bad_kwargs=False, **kwargs):
    """
    Filter a dataframe and adjust its limits.

    Parameters
    ----------
    df
        The input dataframe
    ignore_bad_kwargs
        Ignore kwargs that dont apply to df, else raise.
    kwargs
        The keyword arguments for filtering.
    """
    # apply filtering, this creates a copy so we *should* be ok to update inplace.
    out = df[filter_df(df, ignore_bad_kwargs=ignore_bad_kwargs, **kwargs)]
    # find slice kwargs, get series corresponding to interval columns
    for (name, qs) in yield_slice_from_kwargs(out, kwargs):
        start, stop, step = get_interval_columns(out, name)
        min_val = qs.start if qs.start is not None else start.min()
        max_val = qs.stop if qs.stop is not None else stop.max()
        too_small = start < min_val
        too_large = stop > max_val
        out.loc[too_large, too_large.name] = max_val
        out.loc[too_small, too_small.name] = min_val
    return out


def filter_df(df: pd.DataFrame, ignore_bad_kwargs=False, **kwargs) -> np.array:
    """
    Determine if each row of the index meets some filter requirements.

    Parameters
    ----------
    df
        The input dataframe.
    ignore_bad_kwargs
        If True, silently drop incompatible kwargs with dataframe.

    kwargs
        Used to filter columns.

        Any condition to check against columns of df. Can be a single value
        or a collection of values (to check isin on columns). Str arguments
        can also use unix style matching. Additionally, queries of the form
        {column_name}_min or {column_name}_max can be used, provided columns
        with the same name don't already exist.

    Returns
    -------
    A boolean array of the same len as df indicating if each row meets the
    requirements.
    """
    min_max_query = _convert_times(df, _get_min_max_query(kwargs, df))
    range_query = _add_range_query(kwargs, df, ignore_bad_kwargs)
    multicolumn_range_query = _convert_times(df, range_query)
    equality_query, collection_query = _get_flat_and_collection_queries(kwargs)
    # get a blank index of True for filters
    bool_index = np.ones(len(df), dtype=bool)
    # filter on non-collection queries
    bool_index = _filter_equality(equality_query, df, bool_index)
    # filter on collection queries using isin
    bool_index = _filter_contains(collection_query, df, bool_index)
    # filter based on min/max query
    bool_index = _filter_range(min_max_query, df, bool_index)
    # filter based on ranges
    bool_index = _filter_multicolumn_range(multicolumn_range_query, df, bool_index)
    return bool_index


def _convert_min_max_in_kwargs(kwargs, df):
    """
    Convert the min/max values in kwargs to single key form.

    For example, {'time_min': 10, 'time_max': 20} would be converted
    to {'time': (10, 20)}
    """
    out = dict(kwargs)
    minmax = defaultdict(lambda: [None, None])
    col_set = set(df.columns)
    max_kwargs = {x for x in col_set & set(out) if x.endswith("_max")}
    min_kwargs = {x for x in col_set & set(out) if x.endswith("_min")}
    datetime_cols = set(df.select_dtypes(include=np.datetime64).columns)
    iterable = zip([min_kwargs, max_kwargs], ["_min", "_max"], [0, 1])
    for minmax_kwargs, suffix, ind in iterable:
        for key in minmax_kwargs:
            val = out.pop(key)
            if key in datetime_cols:
                val = to_datetime64(val)
            minmax[key.replace(suffix, "")][ind] = val
    out.update(minmax)
    return out


def get_dim_names_from_columns(df: pd.DataFrame) -> list[str]:
    """
    Returns the names of columns which represent and range in the dataframe.

    For example, time_min, time_max, d_time would be returned if in dataframe.
    """
    cols = set(df.columns)
    possible_dims = {
        x.replace("_min", "").replace("_max", "").replace("d_", "") for x in cols
    }
    out = {
        x for x in possible_dims if {f"{x}_min", f"{x}_max", f"d_{x}"}.issubset(cols)
    }
    return sorted(out)


def get_column_names_from_dim(dims: Sequence[str]) -> list:
    """
    Get column names from a sequence of dimensions.
    """
    out = []
    for name in dims:
        out.append(f"{name}_min")
        out.append(f"{name}_max")
        out.append(f"d_{name}")
    return out


def fill_defaults_from_pydantic(df, base_model: Type[BaseModel]):
    """
    Fill missing columns in dataframe with defaults from base_model.

    If the missing column has no default value, raise ValueError.

    Parameters
    ----------
    df
        A dataframe
    base_model
        A pydantic BaseModel
    """
    fields = base_model.__fields__
    missing = set(fields) - set(df.columns)
    required = {x for x in missing if fields[x].required}
    if any(required):
        msg = f"Missing required value: {required}"
        raise ValueError(msg)
    fill = {x: fields[x].default for x in missing}
    return df.assign(**fill)


def list_ser_to_str(ser: pd.Series) -> pd.Series:
    """
    Convert a column of str sequences to a string with commas separating values.
    """
    values = [",".join(x) if not isinstance(x, str) else x for x in ser.values]
    return pd.Series(values, index=ser.index, dtype=object)



"""
Tests for creating/registering accessors.
"""


# class TestDFSBasics:
#     """Test the basic dascore namespace functions."""
#
#     def test_namespace_exists(self, terra15_das_array):
#         """Just ensure namespace returns accessor."""
#         assert hasattr(terra15_das_array, "dascore"), "accessor not registered"
#         acc = terra15_das_array.dascore
#         name = trim_by_time.__name__
#         out = getattr(acc, name)
#         assert callable(out)



"""
A spool for working with a single file.
"""

from pathlib import Path
from typing import Optional, Union

from typing_extensions import Self

import dascore as dc
from dascore.constants import SpoolType
from dascore.core.spool import DataFrameSpool
from dascore.io.core import FiberIO


class FileSpool(DataFrameSpool):
    """
    A spool for a single file.

    Parameters
    ----------
    path
        The path to the file.
    file_format
        The format name, optional.
    file_version
        The version string of the format, optional.

    Notes
    -----
    Some file formats support storing multiple patches, this is most useful
    for those formats, but should work on all dascore supported formats.
    """

    def __init__(
        self,
        path: Union[str, Path],
        file_format: Optional[str] = None,
        file_version: Optional[str] = None,
    ):
        super().__init__()
        self._path = Path(path)
        if not self._path.exists() or self._path.is_dir():
            msg = f"{path} does not exist or is a directory"
            raise FileNotFoundError(msg)

        _format, _version = dc.get_format(path, file_format, file_version)
        source_df = dc.scan_to_df(path, file_format=_format, file_version=_version)
        dfs = self._get_dummy_dataframes(source_df)
        self._df, self._source_df, self._instruction_df = dfs
        self._file_format = _format
        self._file_version = _version

    def __str__(self):
        """Returns a (hopefully) useful string rep of spool."""
        out = f"FileSpool object managing {self._path}"
        return out

    __repr__ = __str__

    def _load_patch(self, kwargs) -> Self:
        """Given a row from the managed dataframe, return a patch."""
        return dc.read(**kwargs)[0]

    def update(self: SpoolType) -> Self:
        """
        Update the spool.

        If the file format supports indexing (e.g. DASDAE) this will
        trigger an indexing of the file.
        """
        formater = FiberIO.manager.get_fiberio(self._file_format, self._file_version)
        getattr(formater, "index", lambda x: None)(self._path)
        return self



"""Utilities for Quantx."""

from typing import Optional

import numpy as np

from dascore.constants import timeable_types
from dascore.core import Patch
from dascore.core.schema import PatchFileSummary
from dascore.utils.misc import get_slice
from dascore.utils.time import to_datetime64

# --- Getting format/version


def _get_qunatx_version_str(hdf_fi) -> str:
    """
    Return the version string for Quantx file.
    """

    # define a few root attrs that act as a "fingerprint" for Quantx files
    expected_attrs = [
        "GaugeLength",
        "VendorCode",
        "MeasurementStartTime",
        "schemaVersion",
    ]

    acquisition_group = hdf_fi.get_node("/Acquisition")
    acquisition_attrs = acquisition_group._v_attrs

    is_quantx = all([hasattr(acquisition_attrs, x) for x in expected_attrs])
    has_optasense = getattr(acquisition_attrs, "VendorCode", None)
    if not (is_quantx and has_optasense):
        return ""
    return str(acquisition_attrs.schemaVersion.decode())


# --- Getting File summaries


def _get_scanned_time_min_max(data_node):
    """Get the min/max time from time array."""
    # time array is an int with microseconds as precision, so it is better
    # to not convert to float and potentially lose precision.
    time = data_node["RawDataTime"]
    t_len = len(time)
    # first try fast path by tacking first/last of time
    # Note: due to numpy #22197 we have to cast to python int
    tmin = np.datetime64(int(time[0]), "us")
    tmax = np.datetime64(int(time[-1]), "us")
    assert tmin < tmax, "Bad timeblock found, contact the developers."
    return tmin, tmax, t_len


def _get_extra_scan_attrs(self, file_version, path, data_node):
    """Get the extra attributes that go into summary information."""
    tmin, tmax, _ = _get_scanned_time_min_max(data_node)
    out = {
        "time_min": to_datetime64(tmin),
        "time_max": to_datetime64(tmax),
        "path": path,
        "file_format": self.name,
        "file_version": str(file_version),
    }
    return out


def _get_version_data_node(root):
    """Get the version, time, and data node from Quantx file."""
    version = str(root._v_attrs.schemaVersion.decode())
    if version == "2.0":
        data_type = "Raw[0]"
        data_node = root[data_type]
    else:
        raise NotImplementedError("Unknown Quantx version")
    return version, data_node


def _scan_quantx(self, fi, path):
    """Scan an Quantx file, return metadata."""
    root = fi.get_node("/Acquisition")
    root_attrs = root._v_attrs
    version, data_node = _get_version_data_node(root)
    out = _get_default_attrs(data_node, root_attrs)
    out.update(_get_extra_scan_attrs(self, version, path, data_node))
    return [PatchFileSummary.parse_obj(out)]


#
# --- Reading patch


def _get_start_stop(time_len, time_lims, file_tmin, dt):
    """Get the start/stop index along time axis."""
    # sst start index
    tmin = time_lims[0] or file_tmin
    tmax = time_lims[1] or dt * (time_len - 1) + file_tmin
    start_ind = int(np.round((tmin - file_tmin) / dt))
    stop_ind = int(np.round((tmax - file_tmin) / dt)) + 1
    # enforce upper limit on time end index.
    if stop_ind > time_len:
        stop_ind = time_len
    assert 0 <= start_ind < stop_ind
    return start_ind, stop_ind


def _get_dar_attrs(data_node, root, tar, dar):
    """Get the attributes for the Quantx data array (loaded)"""
    attrs = _get_default_attrs(data_node, root._v_attrs)
    attrs["time_min"] = tar.min()
    attrs["time_max"] = tar.max()
    attrs["distance_min"] = dar.min()
    attrs["distance_max"] = dar.max()
    return attrs


def _get_distance_array(root):
    """
    Return an array of distance along fiber values.
    """
    vattrs = root._v_attrs
    channel_numbers = np.arange(vattrs.NumberOfLoci) + vattrs.StartLocusIndex
    assert vattrs.SpatialSamplingIntervalUnit == b"m", "expecting distance in m"
    return channel_numbers * vattrs.SpatialSamplingInterval


def _read_quantx(
    root,
    time: Optional[tuple[timeable_types, timeable_types]] = None,
    distance: Optional[tuple[float, float]] = None,
) -> Patch:
    """
    Read an Quantx file.
    """
    # get time array
    time_lims = tuple(
        to_datetime64(x) if x is not None else None
        for x in (time if time is not None else (None, None))
    )
    _, data_node = _get_version_data_node(root)
    file_t_min, file_t_max, time_len = _get_scanned_time_min_max(data_node)
    dt = (file_t_max - file_t_min) / (time_len - 1)
    # get the start and stop along the time axis
    start_ind, stop_ind = _get_start_stop(time_len, time_lims, file_t_min, dt)
    req_t_min = file_t_min if start_ind == 0 else file_t_min + dt * start_ind
    # account for files that might not be full, adjust requested max time
    stop_ind = min(stop_ind, time_len)
    assert stop_ind > start_ind
    req_t_max = (
        time_lims[-1] if stop_ind < time_len else file_t_min + (stop_ind - 1) * dt
    )
    assert req_t_max > req_t_min
    # calculate time array, convert to datetime64
    time_ar = to_datetime64(file_t_min + np.arange(start_ind, stop_ind) * dt)
    time_inds = (start_ind, stop_ind)
    # get data and sliced distance coord
    dist_ar = _get_distance_array(root)
    dslice = get_slice(dist_ar, distance)
    dist_ar_trimmed = dist_ar[dslice]
    data = data_node.RawData[slice(*time_inds), dslice]
    coords = {"time": time_ar, "channel_number": dist_ar_trimmed}
    dims = ("time", "channel_number")
    attrs = _get_dar_attrs(data_node, root, time_ar, dist_ar_trimmed)
    return Patch(data=data, coords=coords, attrs=attrs, dims=dims)


def _get_default_attrs(data_node, root_node_attrs):
    """
    Return the required/default attributes which can be fetched from attributes.

    Note: missing time, distance absolute ranges. Downstream functions should handle
    this.
    """
    # breakpoint()
    out = dict(dims="time, channel_number", data_category="DAS")
    _root_attrs = {
        "SpatialSamplingInterval": "d_distance",
        "SpatialSamplingIntervalUnit": "distance_units",
        "uuid": "intrument_id",  # not 100% sure about this one
    }
    for treble_name, out_name in _root_attrs.items():
        out[out_name] = getattr(root_node_attrs, treble_name)
    # get calculated attributes
    timing = data_node.RawDataTime
    out["d_time"] = np.timedelta64(timing[1] - timing[0], "us")
    d_ind_start = root_node_attrs.StartLocusIndex
    d_ind_num = root_node_attrs.NumberOfLoci
    out["distance_min"] = d_ind_start * out["d_distance"]
    out["distance_max"] = (d_ind_start + d_ind_num) * out["d_distance"]
    # get attributes from data node
    out["data_units"] = data_node._v_attrs.RawDataUnit.decode()
    # when radians are part of the units this is a phase measurement
    if "rad" in out["data_units"]:
        out["data_type"] = "phase"
    return out



"""
Module to transform a Patch into spectrograms.
"""

from scipy.signal import spectrogram as scipy_spectrogram

from dascore.constants import PatchType
from dascore.utils.misc import _get_sampling_rate
from dascore.utils.patch import patch_function
from dascore.utils.time import to_timedelta64
from dascore.utils.transformatter import FourierTransformatter


def _get_new_attrs_coords(patch, new_dims, axis, new_array, old_array):
    """Get new attributes and coords for spectrogram transform."""
    new_name = new_dims[axis]
    old_name = new_dims[-1]
    coords = dict(patch.coords)
    coords[new_name] = new_array
    coords[old_name] = old_array
    return dict(patch.attrs), coords


@patch_function()
def spectrogram(patch: PatchType, dim: str = "time", **kwargs) -> PatchType:
    """
    Convert a patch to spectrogram arrays along a desired dimension.

    Parameters
    ----------
    patch
        If not None, an axis on which to plot.
    dim
        The dimension along which the spectrograms are calculated.
    **kwargs
        Passed directly to `scipy.signal.spectrogram` to control spectrogram
        construction.
    """
    ft = FourierTransformatter()
    assert dim == "time", "only supporting time for now."
    axis = patch.dims.index(dim)
    fs = _get_sampling_rate(patch.attrs[f"d_{dim}"])
    # returns frequency, new values for original dimension (eg time) and spectrogram
    freqs, original, spec = scipy_spectrogram(patch.data, fs=fs, axis=axis, **kwargs)
    if dim == "time":
        original = to_timedelta64(original)
    new_coord = original + patch.attrs[f"{dim}_min"]
    new_dims = list(ft.rename_dims(patch.dims, index=axis)) + [dim]
    attrs, coords = _get_new_attrs_coords(patch, new_dims, axis, freqs, new_coord)
    return patch.__class__(spec, coords=coords, dims=new_dims, attrs=patch.attrs)



"""
Create the html tables for parameters and such from dataframes.
"""
import hashlib
import inspect
import json
import os
import typing
from functools import cache
from pathlib import Path

import pandas as pd
from jinja2 import Environment, FileSystemLoader

RENDER_FUNCS = {}
DOC_PATH = Path(__file__).absolute().parent.parent / "docs"
API_DOC_PATH = DOC_PATH / "api"
API_DOC_PATH.mkdir(exist_ok=True, parents=True)
TEMPLATE_PATH = DOC_PATH / "_templates"
GITHUB_PATH = "https://github.com"
GITHUB_REPOSITORY = "DASDAE/dascore"
GITHUB_REF = "/master"


@cache
def get_env():
    """Get the template environment."""
    template_path = Path(__file__).absolute().parent / "_templates"
    env = Environment(loader=FileSystemLoader(template_path))
    return env


@cache
def get_template(name):
    """Get the template for rendering tables."""
    env = get_env()
    template = env.get_template(name)
    return template


def _simple_plural(text):
    """return plural form of string"""
    if text.endswith("s"):
        return text + "es"
    return text + "s"


def sha_256(path_or_str: Path | str) -> str:
    """Get the Sha256 hash of a file or a string."""
    if isinstance(path_or_str, Path) and Path(path_or_str).exists():
        with open(path_or_str) as fi:
            path_or_str = fi.read()
    return hashlib.sha256(str(path_or_str).encode('utf-8')).hexdigest()


def build_table(df: pd.DataFrame, caption=None):
    """
    An opinionated function to make a dataframe into html table.
    """
    df = df.drop_duplicates()
    template = get_template("table.html")
    columns = [x.capitalize() for x in df.columns]
    rows = df.to_records(index=False).tolist()
    out = template.render(columns=columns, rows=rows, caption=caption)
    return out


def is_it_subclass(obj, cls):
    """A more intelligent issubclass that doesn't through TypeErrors."""
    try:
        return issubclass(obj, cls)
    except TypeError:
        return False


def _deduplicate_list(seq):
    """Return a list with duplicate removed, preserves order."""
    seen = set()
    return [x for x in seq if not (x in seen or seen.add(x))]


def unpact_annotation(obj, data_dict, address_dict) -> str:
    """Convert an annotation to string with linking"""
    str_id = str(id(obj))
    str_rep = str(obj)
    # this is a basic type, just get its name.
    if is_it_subclass(obj, (str, int, float)):
        out = str_rep.replace("<class ", "").replace(">", "").replace("'", "")
    # this is a forward ref, strip out name and hope for the best ;)
    elif isinstance(obj, typing.ForwardRef):
        # todo: this needs to be more generic, but forward refs are hard!
        # just assume the bound string is resolvable
        key = str(obj).split("('")[-1].replace("')", "")
        name = key.split(".")[-1]
        return f"[{name}](`{key}`)"
    # Array-like thing from numpy #TODO improve this
    elif 'numpy.typing' in str_rep:
        return 'ArrayLike'
    # this is a union, unpack it
    elif str_rep.startswith("typing.Union") or str_rep.startswith('typing.Optional'):
        annos = [unpact_annotation(x, data_dict, address_dict) for x in obj.__args__]
        out = " | ".join(_deduplicate_list(annos))
    # This is a Dict, List, Tuple, etc. Just unpack.
    elif hasattr(obj, "__args__"):
        name = str_rep.replace("typing.", "").split("[")[0]
        sub = [unpact_annotation(x, data_dict, address_dict) for x in obj.__args__]
        out = f"{name}[{', '.join(sub)}]"
    # this is a typevar
    elif hasattr(obj, "__bound__"):
        out = unpact_annotation(obj.__bound__, data_dict, address_dict)
    # this is a class in dascore
    elif str_id in data_dict:
        data = data_dict[str_id]
        out = f"[{data['name']}](`{data['key']}`)"
    # this is a typing type, like Any
    elif str_rep.startswith("typing."):
        out = str_rep.split("typing.")[-1]
    # NoneType should just be None
    elif is_it_subclass(obj, type(None)):
        out = "None"
    # ...
    elif obj is Ellipsis:
        out = "..."
    # generic class from another library
    elif str_rep.startswith("<class"):
        out = str_rep.split(".")[-1].replace("'>", "")
    # return a self type.
    elif str_rep.endswith('.Self'):
        out = 'Self'
    # probably a literal, just give up here
    else:
        if isinstance(obj, str):  # make str look like str
            out = f"'{str(obj)}'"
        else:
            out = str(obj)
    return out.strip()


def build_signature(data, data_dict, address_dict):
    """Return html of signature block."""

    def get_annotation_str(param):
        """Get string of annotation."""
        if not param:
            return ""
        annotation_str = unpact_annotation(param, data_dict, address_dict)
        return f": {annotation_str}"

    def get_param_prefix(kind):
        """Get the prefix for a parameter (eg ** in **kwargs)"""
        if kind == inspect.Parameter.VAR_POSITIONAL:
            return '*'
        elif kind == inspect.Parameter.VAR_KEYWORD:
            return '**'
        return ''

    def get_default_value(param, sig):
        """Get the default value for a parameter."""
        default = sig.parameters[param].default
        if default is inspect._empty:
            return
        return str(default)

    def get_params(sig, annotations):
        out = []
        for param, value in sig.parameters.items():
            prefix = get_param_prefix(value.kind)
            annotation_str = get_annotation_str(annotations.get(param))
            return_str = get_default_value(param, sig)
            param_str = f"{prefix + param}{annotation_str}\n"
            if return_str is not None:
                param_str += f" = {return_str}"
            out.append(param_str)
        return out

    def get_return_line(sig):
        annotation = sig.return_annotation
        if annotation is inspect._empty:
            return_str = ")"
        else:
            return_str = unpact_annotation(annotation, data_dict, address_dict)
            return_str = f")-> {return_str}"
        return return_str

    def get_sig_dict(data):
        """Create a dict of render-able signature stuff."""
        sig = data["signature"]
        annotations = typing.get_type_hints(data["object"])
        out = dict(
            params=get_params(sig, annotations),
            return_line=get_return_line(sig),
            name=data["name"],
        )
        return out

    # no need to do anything if entity is not callable.
    if not data["signature"]:
        return ""

    template = get_template("signature.html")
    sig_dict = get_sig_dict(data)
    out = template.render(**sig_dict)
    return out


class NumpyDocStrParser:
    """Class which parses/process docstrings written in numpy style."""

    heading_char = "##"

    def __init__(self, data):
        self._data = data

    def parse_sections(self, docstr):
        out = {}
        doc_lines = docstr.split("\n")
        dividers = (
                [0]
                + [
                    num
                    for num, x in enumerate(doc_lines)
                    if (set(x).issubset({" ", "-", "\t"}) and ("-" in x))
                ]
                + [len(doc_lines) + 2]
        )
        for start, stop in zip(dividers[:-1], dividers[1:]):
            # determine header or just use 'pre' for txt before headings
            header = "pre" if start == 0 else doc_lines[start - 1].strip()
            # skips the ----- line where applicable
            lstart = start if header == "pre" else start + 1
            out[header] = "\n".join(doc_lines[lstart: stop - 2]).strip()
        return out

    def style_parameters(self, param_str):
        """Style the parameters block."""
        lines = param_str.split("\n")
        # parameters dont have spaces at the start
        param_start = [num for num, x in enumerate(lines) if not x.startswith(" ")]
        param_start.append(len(lines))
        # parse parameters into (parameter): txt
        param_desc = []
        for ind_num, ind in enumerate(param_start[:-1]):
            key = lines[ind].strip()
            vals = [x.strip() for x in lines[ind + 1: param_start[ind_num + 1]]]
            param_desc.append((key, "\n".join(vals)))
        table = pd.DataFrame(param_desc, columns=["Parameter", "Description"])
        return build_table(table)

    def style_examples(self, example_str):
        """
        Styles example blocks.

        Example blocks ban be written in standard doc-test or quarto styles.
        """

        def style_docstest_string(docstr_split):
            """style the doctest string"""
            # need to add comment before lines which dont have comment
            # or dont start with >>> (these are outputs)
            out = []
            for line in docstr_split:
                # the [1:] for > and . lines is to strip off the first space
                if line.startswith(">"):
                    out.append(line.strip(">")[1:])
                elif line.startswith(".."):
                    out.append(line.strip(".")[1:])
                elif line.startswith("#") or not line.strip():
                    out.append(line)
                else:
                    out.append(f"# {line}")
            out_str = "\n".join(out)
            return "\n```{python}\n" + out_str + "\n```"

        # determine if this is doctest by looking for >>>
        line_split = example_str.split("\n")
        if any(x.startswith(">>") for x in example_str.split("\n")):
            return style_docstest_string(line_split)
        else:
            # just need to make python in brackets for quarto
            return example_str.replace("```python", "```{python}")

    def style_notes(self, notes_str):
        """styles notes section of str."""
        template = get_template("notes.md")
        return template.render(note_text=notes_str)

    def style_sections(self, raw_sections):
        """Apply styling to sections."""
        out = dict(raw_sections)
        lower2upper = {x.lower(): x for x in out}
        overlap = set(lower2upper) & set(self.stylers)
        for name in overlap:
            upper_name = lower2upper[name]
            out[upper_name] = self.stylers[name](self, out[upper_name])
        return out

    def __call__(self):
        """Process the docstring."""
        sections_raw = self.parse_sections(self._data["docstring"])
        sections_styled = self.style_sections(sections_raw)
        out = []
        for name, content in sections_styled.items():
            if name == "pre" or "note" in name.lower():
                out.append(content)
            else:
                out.append(f"\n{self.heading_char} {name}\n {content}")
        return "\n".join(out)

    # stylers is used to decide which style function to apply for various
    # sections.
    stylers = {
        "parameter": style_parameters,
        "parameters": style_parameters,
        "examples": style_examples,
        "example": style_examples,
        "notes": style_notes,
        "note": style_notes,
    }


class Render:
    """A class to render API docs for each entity."""

    # The order in which to render the tables.
    _table_order = ("parameter", "method", "attribute", "function", "class", "module")

    def __init__(self, data_dict, object_id, address_dict):
        self._data_dict = data_dict
        self._data = data_dict[object_id]
        self._address_dict = address_dict

    def has_subsection(self, name):
        """Return True if a module has a subsection."""
        possible_children = self._data.get(name, {})
        children = self.get_children_object_ids(possible_children)
        return bool(children)

    def render_linked_table(self, name):
        """Render a table for a specific name."""
        out = []
        for oid in self._data[name]:
            if oid not in self._data_dict:
                continue
            sub_data = self._data_dict[oid]
            linked_name = f"[{sub_data['name']}](`{sub_data['key']}`)"
            desc = sub_data["short_description"]
            out.append((linked_name, desc))
        df = pd.DataFrame(out, columns=["name", "description"])
        return build_table(df)

    def get_children_object_ids(self, ids):
        """Of an object id list, return the children of the current object."""
        key = self._data["key"]
        out = {
            x for x in ids if x in self._data_dict and key in self._data_dict[x]["key"]
        }
        return out

    def _get_github_source(self, data):
        """Get the github source url."""
        obj = data["object"]
        source_code, line_start = inspect.getsourcelines(obj)
        line_end = line_start + len(source_code)
        rel_path = data["path"].relative_to(data["base_path"])
        # grab repo stuff from environment (on GH actions) or defaults
        github_url = os.environ.get("GITHUB_SERVER_URL", GITHUB_PATH)
        owner_repo = os.environ.get("GITHUB_REPOSITORY", GITHUB_REPOSITORY)
        branch = os.environ.get("GITHUB_REF", GITHUB_REF).split("/")[-1]
        source_url = (
            f"{github_url}/{owner_repo}/blob/{branch}/{str(rel_path)}"
            f"#L{line_start + 1}-L{line_end}"
        )
        return source_url

    def _get_parent_source_block(self, data):
        """Create a parent block with a link."""
        parent = data['key'].removesuffix(f".{data['name']}")
        if parent:
            parent_str = f" of [{parent}](`{parent}`)"
        else:
            parent_str = ""
        origin_txt = f"*{self._data['data_type']}* {parent_str}"
        source_url = self._get_github_source(data)
        template = get_template("parent_source_block.html")
        out = template.render(origin_txt=origin_txt, source_url=source_url)
        return out

    def render_markdown(self, heading="##"):
        """Convert numpy docstrings to markdown with some html styling."""
        data = self._data
        docstr = NumpyDocStrParser(data)

        tables = [
            f"\n{heading} {_simple_plural(x).capitalize()}\n"
            f"{self.render_linked_table(x)}\n"
            for x in self._table_order
            if self.has_subsection(x)
        ]
        signature = build_signature(data, self._data_dict, self._address_dict)
        out = (
            f"# {self._data['name']}\n\n"
            f"{self._get_parent_source_block(data)}\n\n"
            f"{signature}\n"
            f"{docstr()}\n"
            f"{''.join(tables)}\n"
        )
        return out


def create_json_mapping(data_dict, obj_dict, api_path):
    """Create the mapping which links address to file path."""
    out = {}

    # add code paths
    for obj_id, data in data_dict.items():
        sub_dir = api_path / "/".join(data["key"].split(".")[:-1])
        path = api_path / sub_dir / f"{data['name']}.qmd"
        out[data["key"]] = f"/{path.relative_to(DOC_PATH)}"

    # add alias (eg import shortcuts/imports from other modules)
    for alias in set(obj_dict) - set(out):
        obj_id = obj_dict[alias]
        if obj_id not in data_dict:
            continue
        main_key = data_dict[obj_id]["key"]
        out[alias] = out[main_key]
    return out


def write_api_markdown(data_dict, api_path, address_dict, debug=False):
    """write all the markdown to disk."""
    files_to_delete = set(Path(api_path).rglob('*.qmd'))
    for obj_id, data in data_dict.items():
        # get path and ensure parents exist
        sub_dir = api_path / "/".join(data["key"].split(".")[:-1])
        path = api_path / sub_dir / f"{data['name']}.qmd"
        path.parent.mkdir(exist_ok=True, parents=True)
        # remove path from files to delete
        if path in files_to_delete:
            files_to_delete.remove(path)
        # dont render non-target file if debugging
        if debug and data['name'] != 'read':
            continue
        # render and write
        render = Render(data_dict, obj_id, address_dict)
        markdown = render.render_markdown()
        # check if file has changed, if not don't write
        if path.exists() and sha_256(path) == sha_256(markdown):
            continue
        path.write_text(markdown)
    # remove files that are no longer writen. This can happen when the code is
    # refactored or objects are deleted.
    for path in files_to_delete:
        path.unlink()


def render_project(data_dict, address_dict, api_path=API_DOC_PATH, debug=False):
    """Render the markdown files."""
    # Create and write the qmd files for each function/class/module
    write_api_markdown(data_dict, api_path, address_dict, debug=debug)
    # dump the json mapping to disk in doc folder
    path_mapping = create_json_mapping(data_dict, address_dict, api_path)
    with open(Path(api_path) / "cross_ref.json", "w") as fi:
        json.dump(path_mapping, fi, indent=2)



"""
A script to create an index of the structure of the DASCore package.
"""
from __future__ import annotations

import inspect
import os
from collections import defaultdict
from importlib import import_module
from pathlib import Path
from types import MethodType, ModuleType, FunctionType
from typing import Literal


def _unwrap_obj(obj):
    """Unwrap a decorated object."""
    while getattr(obj, "__wrapped__", None) is not None:
        obj = obj.__wrapped__
    return obj


def _get_file_path(obj):
    """Try to get the file of a python object."""
    obj = _unwrap_obj(obj)
    try:
        path = inspect.getfile(obj)
    except TypeError:
        path = ""
    return Path(path)


def _get_base_address(path, base_path):
    """
    Get the base address inherent in the path.

    For example, dascore/core/patch.py should return

    dascore.core.patch
    """
    try:
        out = Path(path).relative_to(Path(base_path))
    except ValueError:
        return ""
    new = str(out).replace("/__init__.py", "").replace(".py", "")
    return new.replace("/", ".")


def parse_project(obj, key=None):
    """Parse the project create dict of data and data_type"""

    def yield_get_submodules(obj, base_path):
        """Dynamically load submodules that may not have been imported."""
        path = Path(inspect.getfile(obj))
        if not isinstance(obj, ModuleType) or path.name != "__init__.py":
            return
        submodules = path.parent.glob("*")
        for submod_path in submodules:
            is_dir = submod_path.is_dir()
            is_init = submod_path.name.endswith("__init__.py")
            # this is a directory, look for corresponding __init__.py
            if is_dir and (submod_path / "__init__.py").exists():
                mod_name = str(submod_path.relative_to(base_path)).replace(os.sep, ".")
                mod = import_module(mod_name)
                yield mod_name, mod
            elif submod_path.name.endswith(".py") and not is_init:
                mod_name = (
                    str(submod_path.relative_to(base_path))
                    .replace(".py", "")
                    .replace(os.sep, ".")
                )
                mod = import_module(mod_name)
                yield mod_name, mod

    def get_type(
        obj, parent_is_class=False
    ) -> None | Literal["module", "function", "method", "class"]:
        """Return a string of the type of object."""
        obj = _unwrap_obj(obj)
        if isinstance(obj, ModuleType):
            return "module"
        elif isinstance(obj, MethodType):
            return "method"
        elif isinstance(obj, FunctionType):
            # since this is a class not instance we need this check.
            if parent_is_class:
                return "method"
            else:
                return "function"
        elif isinstance(obj, type):
            return "class"
        return None

    def extract_data(obj, parent_is_class):
        """
        Make lists of attributes, methods, etc.

        These can be feed to render templates.
        """

        try:
            sig = inspect.signature(obj)
        except (TypeError, ValueError):
            sig = None

        docstr = inspect.getdoc(obj)
        dtype = get_type(obj, parent_is_class)
        data = defaultdict(list)
        data["docstring"] = docstr
        data["signature"] = sig
        data["data_type"] = dtype
        data["short_description"] = docstr.split("\n")[0]
        data["object"] = obj
        # get sub-motdules, methods, functions, etc. (just one level deep)

        subs = list(inspect.getmembers(obj)) + list(
            yield_get_submodules(obj, base_path)
        )
        for name, sub_obj in subs:
            if name.startswith("_"):
                continue
            sub_obj = _unwrap_obj(sub_obj)
            sub_dtype = get_type(sub_obj, dtype == "class")
            # for modules, skip entities that aren't children
            if dtype == 'module':
                path, sub_path = _get_file_path(obj), _get_file_path(sub_obj)
                if not str(path).replace('/__init__.py', '') in str(sub_path):
                    continue
            data[sub_dtype].append(str(id(sub_obj)))

        return data

    def get_data(obj, key, base_path, parent_is_class):
        """Get data from object."""
        path = inspect.getfile(obj)
        base_address = _get_base_address(path, base_path)
        data = extract_data(obj, parent_is_class)
        data["base_path"] = Path(base_path)
        data["path"] = Path(path)
        data["key"] = key
        data["name"] = key.split(".")[-1]
        data["base_address"] = base_address
        return data

    def get_address(obj, rel_path):
        """Get the address for an object."""
        base_list = (
            str(rel_path)
            .replace(f"{os.sep}__init__.py", "")
            .replace(".py", "")
            .split(os.sep)
        )
        if not isinstance(obj, ModuleType) and hasattr(obj, "__name__"):
            name_list = [obj.__name__]
        else:
            name_list = []
        return ".".join(base_list + name_list)

    def traverse(obj, data_dict, base_path, key=None, parent_is_class=False):
        """Traverse tree, populate data_dict"""
        obj = _unwrap_obj(obj)
        obj_id = str(id(obj))
        path = _get_file_path(obj)
        # this is something outside of dascore
        if str(base_path) not in str(path) or obj_id in data_dict:
            return
        # load all the modules first
        if isinstance(obj, ModuleType):
            key = get_address(obj, path.relative_to(base_path))
            data_dict[obj_id] = get_data(obj, key, base_path, parent_is_class)
            for _, mod in yield_get_submodules(obj, base_path):
                traverse(mod, data_dict, base_path)
            for name, obj in inspect.getmembers(obj):
                # recurse non-private methods
                if not name.startswith("_"):
                    traverse(obj, data_dict, base_path, f"{key}.{name}", False)
        # then handle non-modules
        else:
            data_dict[str(id(obj))] = get_data(obj, key, base_path, parent_is_class)
            # recurse attributes and methods of classes
            if inspect.isclass(obj):
                for sub_name, sub_obj in inspect.getmembers(obj):
                    if sub_name.startswith("_") or not callable(sub_obj):
                        continue
                    sub_path = _get_file_path(sub_obj)
                    if not str(base_path) in str(sub_path):
                        continue
                    sub_key = f"{key}.{sub_name}"
                    # make sure this is where the method is defined else skip
                    if sub_path != path:
                        continue
                    traverse(sub_obj, data_dict, base_path, sub_key, True)

    # key = key or getattr(obj, "__name__", None)
    base_path = Path(_get_file_path(obj)).parent.parent
    data_dict = {}
    traverse(obj, data_dict, base_path)
    # traverse(obj, key, data_dict, str(base_path), parent_path=base_path)
    return data_dict


def get_alias_mapping(module, key=None):
    """Return a dict of {object_path: id} to construct cross refs."""

    def traverse_simple(obj, key, data_dict, base_path):
        """Traverse the tree and write out markdown."""
        obj = _unwrap_obj(obj)
        obj_id = str(id(obj))
        path = _get_file_path(obj)
        if not _get_base_address(path, base_path):
            return
        data_dict[key] = obj_id
        for (member_name, member) in inspect.getmembers(obj):
            if member_name.startswith("_"):
                continue
            new_key = ".".join([key, member_name])
            if new_key.split(".")[-1] in key:
                continue
            traverse_simple(member, new_key, data_dict, base_path)

    data_dict = {}
    key = key or getattr(module, "__name__", None)
    base_path = Path(_get_file_path(module)).parent.parent
    traverse_simple(module, key, data_dict, base_path)
    return data_dict



"""
Tests for filters.
"""
import numpy as np
import pandas as pd
import pytest

import dascore
from dascore.exceptions import FilterValueError


class TestFilterChecks:
    """Test that bad filter checks raise appropriate errors."""

    def test_no_kwargs_raises(self, random_patch):
        """Ensure ValueError is raised with no dim parameters."""
        with pytest.raises(FilterValueError):
            _ = random_patch.pass_filter()

    def test_wrong_kwarg_length_raises(self, random_patch):
        """only len two sequence is allowed."""
        with pytest.raises(FilterValueError, match="length two sequence"):
            _ = random_patch.pass_filter(time=[1])
        with pytest.raises(FilterValueError, match="length two sequence"):
            _ = random_patch.pass_filter(time=[1, 3, 3])

    def test_all_null_kwarg_raises(self, random_patch):
        """There must be one Non-null kwarg."""
        with pytest.raises(FilterValueError, match="at least one filter"):
            _ = random_patch.pass_filter(time=[None, np.NAN])

    def test_unordered_params(self, random_patch):
        """Ensure a low parameter greater than a high parameter raises."""
        with pytest.raises(FilterValueError):
            _ = random_patch.pass_filter(time=[10, 1])

    def test_bad_low_param(self, random_patch):
        """Ensure a low parameter above niquest raises."""
        with pytest.raises(FilterValueError, match="possible filter bounds"):
            _ = random_patch.pass_filter(time=[1000, None])

    def test_bad_high_param(self, random_patch):
        """Ensure a high parameter above niquest raises."""
        with pytest.raises(FilterValueError, match="possible filter bounds"):
            _ = random_patch.pass_filter(distance=[None, 10])

    def test_filt_param_less_than_0(self, random_patch):
        """Ensure a high parameter above niquest raises."""
        with pytest.raises(FilterValueError, match="possible filter bounds"):
            _ = random_patch.pass_filter(distance=[None, -10])


class TestFilterBasics:
    """Simple tests to make sure filter logic runs."""

    def test_time_bandpass_runs(self, random_patch):
        """Ensure filtering along time_axis works"""
        out = random_patch.pass_filter(time=(10, 100))
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_time_lowpass_runs(self, random_patch):
        """Ensure stopfilter on time works."""
        out = random_patch.pass_filter(time=(None, 100.22))
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_time_highpass_runs(self, random_patch):
        """Ensure a highpass on time runs"""
        out = random_patch.pass_filter(time=(10.2, None))
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_apply_distance_filter(self, random_patch):
        """Apply bandpass along distance dimension."""
        out = random_patch.pass_filter(distance=(0.1, 0.2))
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))


class TestSobelFilter:
    """Simple tests to make sure Sobel filter runs."""

    def test_invalid_mode(self, random_patch):
        """Ensure ValueError is raised with an invalid mode."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim="time", mode="test", cval=0.0)

    def test_invalid_mode_type(self, random_patch):
        """Ensure ValueError is raised with an invalid mode type."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim="time", mode=0.0)

    def test_invalid_dim(self, random_patch):
        """Ensure ValueError is raised with an invalid dim type."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim=-1)

    def test_invalid_axis(self, random_patch):
        """Ensure ValueError is raised with an invalid axis value."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim=None, mode="constant", cval=0.0)

    def test_invalid_cval(self, random_patch):
        """Ensure ValueError is raised with an invalid cval value."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim="distance", mode="constant", cval=None)

    def test_sobel_runs(self, random_patch):
        """Ensure Sobel filter works with default params."""
        out = random_patch.sobel_filter(dim="time")
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))


class TestMedianFilter:
    """Simple tests on median filter"""

    def test_median_filter_default(self, random_patch):
        """apply default values"""
        out = random_patch.median_filter()
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_median_filter_user_single_value(self, random_patch):
        """apply default values"""
        out = random_patch.median_filter(kernel_size=5)
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_median_filter_user_multi_value(self, random_patch):
        """apply default values"""
        out = random_patch.median_filter(kernel_size=(5, 3))
        assert isinstance(out, dascore.Patch)
        assert not np.any(pd.isnull(out.data))



"""An HDF5-based indexer for local file systems."""
import abc
import os
import time
from pathlib import Path
from typing import Optional, Union

import pandas as pd
from typing_extensions import Self

import dascore as dc
from dascore.constants import ONE_SECOND_IN_NS, path_types
from dascore.exceptions import InvalidIndexVersionError
from dascore.utils.hdf5 import HDFPatchIndexManager
from dascore.utils.misc import iter_files, iterate
from dascore.utils.pd import filter_df
from dascore.utils.progress import track
from dascore.utils.time import get_max_min_times, to_timedelta64

# supported read_hdf5 kwargs
READ_HDF5_KWARGS = frozenset(
    {"columns", "where", "mode", "errors", "start", "stop", "key", "chunksize"}
)


class AbstractIndexer:
    """
    A base class for indexers.

    This is primarily here for a place-holder.
    """

    path: Path

    @abc.abstractmethod
    def update(self) -> Self:
        """
        Updates the contents of the Indexer.

        Resets any previous selection.
        """


class DirectoryIndexer(AbstractIndexer):
    """
    A class for indexing a directory of dascore-readable files.

    This works by crawling the directory, getting a summary about the data it
    contains, then creating a small HDF index file which can be queried later
    on.

    Parameters
    ----------
    path
        The path to a directory containing DAS files.
    cache_size
        The number of queries to store in memory to avoid frequent reads
        of the index file. It is rare this needs to be modified.
    index_path
        The path to the index. By default, the index will be created on the
        top level of the data directory.
    """

    # hdf5 compression defaults
    ext = ""
    namespace = ""
    index_name = ".dascore_index.h5"  # name of index file
    executor = None  # an executor for using parallelism

    def __init__(self, path: Union[str, Path], cache_size: int = 5, index_path=None):
        self.max_size = cache_size
        self.path = Path(path).absolute()
        self.index_path = Path(index_path or self.path / self.index_name)
        self.cache = pd.DataFrame(
            index=range(cache_size), columns="t1 t2 kwargs cindex".split()
        )
        self._current_index = 0
        self._index_table = HDFPatchIndexManager(
            self.index_path,
            self.namespace,
        )

    def get_contents(self, buffer=ONE_SECOND_IN_NS, **kwargs) -> pd.DataFrame:
        """
        Get contents of directory with specific query params.

        Parameters
        ----------
        buffer
            A buffer to ensure enough info is returned from hdf index.
        kwargs
            Used to query contents.
        """
        # create index if it doesn't exist
        if not self.index_path.exists():
            self.update()
        # if the index still doesn't exist there are no readable files, return
        # empty df.
        if not self.index_path.exists():
            return pd.DataFrame(columns=self._index_table.index_columns)
        time_min, time_max = get_max_min_times(kwargs.pop("time", None))
        hdf5_kwargs, kwargs = self._separate_hdf5_kwargs(kwargs)
        buffer = to_timedelta64(buffer)
        # find out if the query falls within one cached times
        con1 = self.cache.t1 <= time_min
        con2 = self.cache.t2 >= time_max
        con3 = self.cache.kwargs == self._kwargs_to_str(kwargs)
        cached_index = self.cache[con1 & con2 & con3]
        if not len(cached_index):  # query is not cached get it from hdf5 file
            index = self._index_table.get_index(
                time_min=time_min,
                time_max=time_max,
                **hdf5_kwargs,
            )
            self._set_cache(index, time_min, time_max, hdf5_kwargs)
        else:
            index = cached_index.iloc[0]["cindex"]
        # trim down index
        con1 = index["time_min"] >= (time_max + buffer)
        con2 = index["time_max"] <= (time_min - buffer)
        pre_filter_df = index[~(con1 | con2)]
        out = pre_filter_df[
            filter_df(pre_filter_df, time=(time_min, time_max), **kwargs)
        ]
        return out

    def __str__(self):
        """Rep. indexer as a string."""
        msg = f"{self.__class__.__name__} managing: {self.path}"
        return msg

    __repr__ = __str__

    __call__ = get_contents

    def _separate_hdf5_kwargs(self, kwargs):
        """Ensure kwargs are supported."""
        kdf_kwargs = {i: v for i, v in kwargs.items() if i in READ_HDF5_KWARGS}
        kwargs = {i: v for i, v in kwargs.items() if i not in READ_HDF5_KWARGS}
        return kdf_kwargs, kwargs

    def _set_cache(self, index, starttime, endtime, kwargs):
        """Cache the current index."""
        ser = pd.Series(
            {
                "t1": starttime,
                "t2": endtime,
                "cindex": index,
                "kwargs": self._kwargs_to_str(kwargs),
            }
        )
        self.cache.loc[self._get_next_index()] = ser

    def _kwargs_to_str(self, kwargs):
        """Convert kwargs to a string."""
        keys = sorted(list(kwargs.keys()))
        out = str([(item, kwargs[item]) for item in keys])
        return out

    def clear_cache(self):
        """Removes all cached dataframes."""
        self.cache = pd.DataFrame(
            index=range(self.max_size), columns="t1 t2 kwargs cindex".split()
        )

    def _get_next_index(self):
        """
        Get the next index value on cache.

        Note we can't use itertools.cycle here because it cant be pickled.
        """
        if self._current_index == len(self.cache.index) - 1:
            self._current_index = 0
        else:
            self._current_index += 1
        return self.cache.index[self._current_index]

    def _get_file_iterator(self, paths: Optional[path_types] = None, only_new=True):
        """Return an iterator of potential un-indexed files."""
        # get mtime, subtract a bit to avoid odd bugs
        mtime = None
        # getting last updated might need the db so only call once.
        last_updated = self._index_table.last_updated_timestamp if only_new else None
        if last_updated is not None and only_new:
            mtime = last_updated - 0.001
        # get paths to iterate
        path = self.path
        if paths is None:
            paths = path
        else:
            paths = [
                f"{path}/{x}" if str(path) not in str(x) else str(x)
                for x in iterate(paths)
            ]
        # return file iterator
        return iter_files(paths, ext=self.ext, mtime=mtime)

    def ensure_path_exists(self, create=False):
        """
        Ensure the base path exists else raise.

        If create is True, simply create an empty directory.
        """
        path = Path(self.path)
        if create:
            path.mkdir(parents=True, exist_ok=True)
        if not path.is_dir():
            msg = f"{path} is not a directory, cant read spool"
            raise FileExistsError(msg)

    def _enforce_min_version(self):
        """
        Ensure the minimum version is met, else delete index file.
        """
        try:
            self._index_table.validate_version()
        except InvalidIndexVersionError:
            os.remove(self.index_path)
            self.update()

    def update(self, paths=None) -> Self:
        """
        Updates the contents of the Indexer.

        Resets any previous selection.
        """
        self._enforce_min_version()  # delete index if schema has changed
        update_time = time.time()
        new_files = list(self._get_file_iterator(paths=paths, only_new=True))
        smooth_iterator = track(new_files, f"Indexing {self.path.name}")
        data_list = [y.dict() for x in smooth_iterator for y in dc.scan(x, ignore=True)]
        df = pd.DataFrame(data_list)
        if not df.empty:
            # ensure the base path is not in the path column
            assert "path" in set(df.columns), f"{df} has no path column"
            self._index_table.write_update(df, update_time, base_path=self.path)
            # clear cache out when new traces are added
            self.clear_cache()
        return self



"""
Function for querying Patchs
"""
from dascore.constants import PatchType
from dascore.utils.patch import _AttrsCoordsMixer, _get_history_str, patch_function
from dascore.utils.time import get_select_time


def _prepare_kwargs(patch, kwargs):
    """Maybe convert times to datetime, get min/max values."""
    if "time" in kwargs:
        tmin = patch.attrs.time_min
        tmax = patch.attrs.time_max
        times = tuple(
            get_select_time(x, tmin, tmax) if x is not None else x
            for x in kwargs["time"]
        )
        kwargs["time"] = times
    # convert (start, end) to slice(start, end)
    kwargs = {
        i: slice(v[0], v[1]) if isinstance(v, (list, tuple)) else v
        for i, v in kwargs.items()
    }
    return kwargs


@patch_function(history=None)
def select(patch: PatchType, *, copy=False, **kwargs) -> PatchType:
    """
    Return a subset of the trace based on query parameters.

    Any dimension of the data can be passed as key, and the values
    should either be a Slice or a tuple of (min, max) for that
    dimension.

    The time dimension is handled specially in that either floats,
    datetime64 or datetime objects can be used to specify relative
    or absolute times, respectively.

    Parameters
    ----------
    copy
        If True, copy the resulting data. This is needed so the old
        array can get gc'ed and memory freed.
    **kwargs
        Used to specify the dimension and slices to select on.

    Examples
    --------
    >>> # select meters 50 to 300
    >>> import numpy as np
    >>> from dascore.examples import get_example_patch
    >>> tr = get_example_patch()
    >>> new = tr.select(distance=(50,300))
    """
    kwargs = _prepare_kwargs(patch, kwargs)
    # convert tuples into slices
    new = patch._data_array.sel(**kwargs)
    # no slicing was performed, just return original.
    if new.data.shape == patch.data.shape:
        return patch
    # prepare outputs
    data = new.data if not copy else new.data.copy()
    attrs = dict(new.attrs)
    attrs["history"] = _get_history_str(patch, select, **kwargs)
    attrs, coords = _AttrsCoordsMixer(attrs, new.coords, new.dims)()
    return patch.__class__(data, attrs=attrs, coords=coords, dims=patch.dims)



"""
Misc. tests for misfit utilities.
"""
import os
import time
from pathlib import Path

import numpy as np
import pytest

from dascore.exceptions import MissingOptionalDependency, ParameterError
from dascore.utils.misc import (
    MethodNameSpace,
    check_evenly_sampled,
    get_slice,
    iter_files,
    iterate,
    optional_import,
)


class ParentClass:
    """A test parent class."""

    @property
    def namespace(self):
        """Your run-o-the-mill namespace"""
        return MNS(self)


class MNS(MethodNameSpace):
    """method name space subclass."""

    def func1(self, expected_type):
        """First func"""
        return isinstance(self, expected_type)


class TestNamespaceClass:
    """Tests for namespace class."""

    def test_parent_self_passed_to_namespace(self):
        """Ensure the parent of namespace gets passed to self."""
        pc = ParentClass()
        assert pc.namespace.func1(ParentClass)

    def test_assign_adhoc_method(self):
        """Ensure methods added after class definition still work."""

        def new_method(self, expected_type):
            return isinstance(self, expected_type)

        MNS.new_method = new_method
        pc = ParentClass()
        assert pc.namespace.new_method(ParentClass)


class TestGetSlice:
    """Ensure getting slices of arrays for indexing works."""

    ar = np.arange(100)

    def test_two_intervals(self):
        """test get slice for two intervals"""
        array_slice = get_slice(self.ar, cond=(1, 10))
        expected = slice(1, 11, None)
        assert array_slice == expected

    def test_right_side(self):
        """test for only right interval"""
        array_slice = get_slice(self.ar, cond=(None, 10))
        expected = slice(None, 11, None)
        assert array_slice == expected

    def test_left_side(self):
        """Ensure left side interval works."""
        array_slice = get_slice(self.ar, cond=(1, None))
        expected = slice(1, None, None)
        assert array_slice == expected

    def test_no_bounds(self):
        """Empty slice should be returned when no bounds specified."""
        array_slice = get_slice(self.ar, cond=(None, None))
        expected = slice(None, None, None)
        assert array_slice == expected

    def test_out_of_bounds(self):
        """When out of bounds, non should be returned."""
        array_slice = get_slice(self.ar, cond=(-100, 1_000))
        expected = slice(None, None, None)
        assert array_slice == expected

    def test_cond_is_none(self):
        """Ensure None is a valid input, returns empty slice"""
        array_slice = get_slice(self.ar, cond=None)
        expected = slice(None, None, None)
        assert array_slice == expected

    def test_slice_end_with_zeros(self):
        """
        Ensure we get a slice without None at the end if arrays are zeroed at end.
        """
        ar = np.arange(100)
        ar[-20:] = 0
        sliced = get_slice(ar, (None, ar.max()))
        assert sliced.stop is not None
        assert sliced.stop == ar.max()

    def test_slice_middle_with_zeros(self):
        """
        Ensure we get a slice without None at the end if arrays are zeroed at end.
        """
        ar = np.arange(100)
        ar[-20:] = 0
        sliced = get_slice(ar, (None, ar.max() - 10))
        assert sliced.stop is not None
        assert ar[sliced].max() == (ar.max() - 10)


class TestCheckEvenlySampled:
    """Tests for uniform sampling check."""

    dt = np.datetime64("2021-01-01")
    td = np.timedelta64(1, "s")

    def test_check_uniform(self):
        """Ensure no exception is raised for uniform sampled arrays."""
        arrays = dict(
            a=[1, 2, 3],
            b=np.arange(0, 10),
            c=[self.dt + self.td, self.dt + 2 * self.td, self.dt + 3 * self.td],
            d=np.linspace(0, 1, 50),
        )
        for _, array in arrays.items():
            check_evenly_sampled(array)

    def test_check_non_uniform(self):
        """Ensure no exception is raised for uniform sampled arrays."""

        arrays = dict(
            a=[1, 2, 3, 5],
            c=[self.dt + self.td, self.dt + 4 * self.td, self.dt + 3 * self.td],
        )
        for _, array in arrays.items():
            with pytest.raises(ParameterError):
                check_evenly_sampled(array)


class TestIterFiles:
    """Tests for iterating directories of files."""

    sub = {"D": {"C": ".mseed"}, "F": ".json", "G": {"H": ".txt"}}
    file_paths = {"A": ".txt", "B": sub}

    # --- helper functions
    def setup_test_directory(self, some_dict: dict, path: Path):
        """Build the test directory."""
        for path in self.get_file_paths(some_dict, path):
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("w") as fi:
                fi.write("useful text")

    def get_file_paths(self, some_dict, path):
        """Return expected paths to files."""
        for i, v in some_dict.items():
            if isinstance(v, dict):
                yield from self.get_file_paths(v, path / i)
            else:
                yield path / (i + v)

    # --- fixtures
    @pytest.fixture(scope="class")
    def simple_dir(self, tmp_path_factory):
        """Return a simple directory for iterating."""
        path = Path(tmp_path_factory.mktemp("iterfiles"))
        self.setup_test_directory(self.file_paths, path)
        return path

    @pytest.fixture(scope="class")
    def dir_with_hidden_dir(self, tmp_path_factory):
        """Create a directory with a hidden directory inside."""
        path = Path(tmp_path_factory.mktemp("iterfiles_hidden"))
        struct = dict(self.file_paths)
        # add hidden directory with files in it.
        struct[".Hidden"] = {"Another": {"hidden_by_parent": ".txt"}}
        self.setup_test_directory(struct, path)
        return path

    def test_basic(self, simple_dir):
        """test basic usage of iterfiles."""
        files = set(self.get_file_paths(self.file_paths, simple_dir))
        out = set((Path(x) for x in iter_files(simple_dir)))
        assert files == out

    def test_one_subdir(self, simple_dir):
        """Test with one sub directory."""
        subdirs = simple_dir / "B" / "D"
        out = set(iter_files(subdirs))
        assert len(out) == 1

    def test_multiple_subdirs(self, simple_dir):
        """Test with multiple sub directories."""
        path1 = simple_dir / "B" / "D"
        path2 = simple_dir / "B" / "G"
        out = {Path(x) for x in iter_files([path1, path2])}
        files = self.get_file_paths(self.file_paths, simple_dir)
        expected = {
            x
            for x in files
            if str(x).startswith(str(path1)) or str(x).startswith(str(path2))
        }
        assert out == expected

    def test_extention(self, simple_dir):
        """Test filtering based on extention."""
        out = set(iter_files(simple_dir, ext=".txt"))
        for val in out:
            assert val.endswith(".txt")

    def test_mtime(self, simple_dir):
        """Test filtering based on modified time"""
        files = list(self.get_file_paths(self.file_paths, simple_dir))
        # set the first file mtime in future
        now = time.time()
        first_file = files[0]
        os.utime(first_file, (now + 10, now + 10))
        # get output make sure it only returned first file
        out = list(iter_files(simple_dir, mtime=now + 5))
        assert len(out) == 1
        assert Path(out[0]) == first_file

    def test_skips_files_in_hidden_directory(self, dir_with_hidden_dir):
        """Hidden directory files should be skipped."""
        out1 = list(iter_files(dir_with_hidden_dir))
        has_hidden_by_parent = ["hidden_by_parent" in x for x in out1]
        assert not any(has_hidden_by_parent)
        # But if skip_hidden is False it should be there
        out2 = list(iter_files(dir_with_hidden_dir, skip_hidden=False))
        has_hidden_by_parent = ["hidden_by_parent" in x for x in out2]
        assert sum(has_hidden_by_parent) == 1


class TestIterate:
    """Test case for iterate."""

    def test_none(self):
        """None should return an empty tuple"""
        assert iterate(None) == tuple()

    def test_object(self):
        """A single object should be returned in a tuple"""
        assert iterate(1) == (1,)

    def test_str(self):
        """A single string object should be returned as a tuple"""
        assert iterate("hey") == ("hey",)


class TestOptionalImport:
    """Ensure the optional import works."""

    def test_import_installed_module(self):
        """Test to ensure an installed module imports."""
        import dascore as dc

        mod = optional_import("dascore")
        assert mod is dc
        sub_mod = optional_import("dascore.core")
        assert sub_mod is dc.core

    def test_missing_module_raises(self):
        """Ensure a module which is missing raises the appropriate Error."""
        with pytest.raises(MissingOptionalDependency, match="boblib4"):
            optional_import("boblib4")



"""
Tests for selecting data from traces.
"""

import numpy as np

import dascore as dc
from dascore.utils.time import to_timedelta64


class TestSelect:
    """Tests for selecting data from Trace."""

    def test_select_by_distance(self, random_patch):
        """
        Ensure distance can be used to filter trace.
        """
        dmin, dmax = 100, 200
        pa = random_patch.select(distance=(dmin, dmax))
        assert pa.data.shape < random_patch.data.shape
        # the attrs should have updated as well
        assert pa.attrs["distance_min"] >= 100
        assert pa.attrs["distance_max"] <= 200

    def test_select_by_absolute_time(self, random_patch):
        """
        Ensure the data can be sub-selected using absolute time.
        """
        shape = random_patch.data.shape
        t1 = random_patch.attrs["time_min"] + np.timedelta64(1, "s")
        t2 = t1 + np.timedelta64(3, "s")

        pa1 = random_patch.select(time=(None, t1))
        assert pa1.attrs["time_max"] <= t1
        assert pa1.data.shape < shape

        pa2 = random_patch.select(time=(t1, None))
        assert pa2.attrs["time_min"] >= t1
        assert pa2.data.shape < shape

        tr3 = random_patch.select(time=(t1, t2))
        assert tr3.attrs["time_min"] >= t1
        assert tr3.attrs["time_max"] <= t2
        assert tr3.data.shape < shape

    def test_select_by_positive_float(self, random_patch):
        """Floats in time dim should usable to reference start of the trace."""
        shape = random_patch.data.shape
        t1 = random_patch.attrs["time_min"]
        pa1 = random_patch.select(time=(1, None))
        expected_start = t1 + to_timedelta64(1)
        assert pa1.attrs["time_min"] <= expected_start
        assert pa1.data.shape < shape

    def test_select_by_negative_float(self, random_patch):
        """Ensure negative floats reference end of trace."""
        shape = random_patch.data.shape
        pa1 = random_patch.select(time=(None, -2))
        expected_end = random_patch.attrs["time_max"] - to_timedelta64(2)
        assert pa1.attrs["time_max"] >= expected_end
        assert pa1.data.shape < shape

    def test_select_distance_leaves_time_attr_unchanged(self, random_patch):
        """Ensure selecting on distance doesn't change time"""
        dist = random_patch.coords["distance"]
        dist_max, dist_mean = np.max(dist), np.mean(dist)
        out = random_patch.select(distance=(dist_mean, dist_max - 1))
        assert out.attrs["time_max"] == out.coords["time"].max()


class TestSelectHistory:
    """Test behavior of history added by select."""

    def test_select_outside_bounds(self, random_patch):
        """Selecting outside the bounds should do nothing."""
        attrs = random_patch.attrs
        dt = dc.to_timedelta64(1)
        time = (attrs["time_min"] - dt, attrs["time_max"] + dt)
        dist = (attrs["distance_min"] - 1, attrs["distance_max"] + 1)
        new = random_patch.select(time=time, distance=dist)
        # if no select performed everything should be identical.
        assert new.equals(random_patch, only_required_attrs=False)



"""
Script to build the API docs for dascore.
"""
import dascore
from _index_api import parse_project, get_alias_mapping
from _render_api import render_project


if __name__ == "__main__":
    data_dict = parse_project(dascore)
    obj_dict = get_alias_mapping(dascore)
    render_project(data_dict, obj_dict, debug=False)



"""Constants used throughout obsplus."""
from pathlib import Path
from typing import TypeVar, Union

import numpy as np
import pandas as pd

import dascore  # NOQA

PatchType = TypeVar("PatchType", bound="dascore.Patch")

SpoolType = TypeVar("SpoolType", bound="dascore.BaseSpool")

# Bump this to force re-downloading of all data file
DATA_VERSION = "0.0.0"

# Types dascore can convert into time representations
timeable_types = Union[int, float, str, np.datetime64, pd.Timestamp]

# Number types
numeric_types = Union[int, float]

# expected DAS attributes
REQUIRED_DAS_ATTRS = ("d_time", "d_distance")

# The smallest value an int64 can rep. (used as NaT by datetime64)
MININT64 = np.iinfo(np.int64).min

# The largest value an int64 can rep
MAXINT64 = np.iinfo(np.int64).max

# types used to represent paths
path_types = Union[str, Path]

# One second in numpy timedelta speak
ONE_SECOND = np.timedelta64(1, "s")

# One nanosecond
ONE_NANOSECOND = np.timedelta64(1, "ns")

# one billion
ONE_BILLION = 1_000_000_000

# One second with a precision of nano seconds
ONE_SECOND_IN_NS = np.timedelta64(ONE_BILLION, "ns")

# Valid strings for "datatype" attribute
VALID_DATA_TYPES = (
    "",  # unspecified
    "velocity",
    "strain_rate",
    "phase",
    "strain",
    "temperature",
    "temperature_gradient",
)

# Valid categories (of instruments)
VALID_DATA_CATEGORIES = ("", "DAS", "DTS", "DSS")


max_lens = {
    "path": 120,
    "file_format": 15,
    "tag": 100,
    "network": 8,
    "station": 8,
    "dims": 40,
    "file_version": 9,
    "cable_id": 50,
    "instrument_id": 50,
}


# Methods FileFormatter needs to support
FILE_FORMATTER_METHODS = ("read", "write", "get_format", "scan")

# A set of attributes which are used in Patch equality checks.
COMPARE_ATTRS = {
    "d_time",
    "d_distance",
    "data_type",
    "category" "time_min",
    "time_max",
    "distance_min",
    "distance_max",
    "instrument_id",
}

# Large and small np.datetime64[ns] (used when defaults are needed)
SMALLDT64 = np.datetime64(MININT64 + 5_000_000_000, "ns")
LARGEDT64 = np.datetime64(MAXINT64 - 5_000_000_000, "ns")

# Required shared attributes to merge patches together
PATCH_MERGE_ATTRS = ("network", "station", "dims", "data_type", "data_category")


# A map from the unit name to the code used in numpy.timedelta64
NUMPY_TIME_UNIT_MAPPING = {
    "hour": "h",
    "minute": "m",
    "second": "s",
    "millisecond": "ms",
    "microsecond": "us",
    "nanosecond": "ns",
    "picosecond": "ps",
    "femtosecond": "fs",
    "attosecond": "as",
    "year": "Y",
    "month": "M",
    "week": "W",
    "day": "D",
}


# A description of basic patch metadata.
basic_summary_attrs = f"""
data_type
    The type of data collected (the meaning of the data). Valid values
    are {VALID_DATA_TYPES}.
data_category
    The category of instrument which recorded the data. Valid values
    are {VALID_DATA_CATEGORIES}.
data_units
    The units in which the data are recorded (e.g., strain_rate).
d_time
    The temporal sample spacing. If the patch is not evenly sampled
    this should be set to `np.timedelta64('NaT')`
time_min
    The time represented by the first sample in the patch.
time_max
    The time represented by the last sample in the patch.
time_units
    The units of the time axis (in most cases should be seconds) or
    specified by datetime64 arrays in time coordinate.
d_distance
    The spatial sampling rate, set to NaN if the patch is not evenly sampled
    in space.
distance_min
    The along-fiber distance of the first channel in the patch.
distance_max
    The along-fiber distance of the last channel in the patch.
d_distance
    The spatial sampling rate, set to NaN if the patch is not evenly sampled
    in space.
distance_units
    The units of the distance axis. In most cases should be 'm'.
instrument_id
    A unique identifier of the instrument.
cable_id
    A Unique identifier of the cable, or composition of cables.
distance_units
    The units of distance, defaults to m.
data_units
    units
category
    The category
network
    The network code an ascii-compatible string up to 2 characters.
station
    The station code an ascii-compatible string up to 5 characters
instrument_id
    The identifier of the instrument.
dims
    A tuple of dimension names in the same order as the data dimensions.
tag
    A custom string up to 100 chars.
station
    A network code (up to 8 chars).
network
    A network code (up to 8 chars).
history
    A list of strings indicating what processing has occurred on the patch.
"""



"""
Tests for QuantXV2 format
"""
import numpy as np
import pytest

import dascore as dc
from dascore.constants import REQUIRED_DAS_ATTRS
from dascore.core.schema import PatchFileSummary
from dascore.io.core import read
from dascore.io.quantx.core import QuantXV2
from dascore.utils.downloader import fetch
from dascore.utils.misc import register_func

PATCH_FIXTURES = []


@pytest.fixture(scope="session")
def quantx_v2_example_path():
    """Return the path to the example QuantXV2 file."""
    out = fetch("opta_sense_quantx_v2.h5")
    assert out.exists()
    return out


@pytest.fixture(scope="session")
@register_func(PATCH_FIXTURES)
def quantx_v2_das_patch(quantx_v2_example_path):
    """Read the QuantXV2 data, return contained DataArray"""
    out = read(quantx_v2_example_path, "quantx")[0]
    attr_time = out.attrs["time_max"]
    coord_time = out.coords["time"].max()
    assert attr_time == coord_time
    return out


class TestReadQuantXV2:
    """Tests for reading the QuantXV2 format."""

    def test_type(self, quantx_v2_das_patch):
        """Ensure the expected type is returned."""
        assert isinstance(quantx_v2_das_patch, dc.Patch)

    def test_attributes(self, quantx_v2_das_patch):
        """Ensure a few of the expected attrs exist in array."""
        attrs = dict(quantx_v2_das_patch.attrs)
        expected_attrs = {"time_min", "time_max", "distance_min", "data_units"}
        assert set(expected_attrs).issubset(set(attrs))

    def test_has_required_attrs(self, quantx_v2_das_patch):
        """Ensure the required das attrs are found."""
        assert set(REQUIRED_DAS_ATTRS).issubset(set(dict(quantx_v2_das_patch.attrs)))

    def test_coord_attr_time_equal(self, quantx_v2_das_patch):
        """The time reported in the attrs and coords should match."""
        attr_time_max = quantx_v2_das_patch.attrs["time_max"]
        coord_time_max = quantx_v2_das_patch.coords["time"].max()
        assert attr_time_max == coord_time_max
        attr_time_min = quantx_v2_das_patch.attrs["time_min"]
        coord_time_min = quantx_v2_das_patch.coords["time"].min()
        assert attr_time_min == coord_time_min

    def test_precision_of_time_array(self, quantx_v2_das_patch):
        """
        Ensure the time array is in ns, not native ms, in order to
        be consistent with other patches.
        """
        time = quantx_v2_das_patch.coords["time"]
        dtype = time.dtype
        assert "[ns]" in str(dtype)

    def test_time_dist_slice(self, quantx_v2_das_patch, quantx_v2_example_path):
        """Ensure slicing distance and time works from read func."""
        time_array = quantx_v2_das_patch.coords["time"]
        dist_array = quantx_v2_das_patch.coords["channel_number"]
        t1, t2 = time_array[10], time_array[40]
        d1, d2 = dist_array[10], dist_array[40]
        patch = QuantXV2().read(
            quantx_v2_example_path, time=(t1, t2), distance=(d1, d2)
        )[0]
        attrs, coords = patch.attrs, patch.coords
        assert attrs["time_min"] == coords["time"].min() == t1
        assert (attrs["time_max"] - t2) < (attrs["d_time"] / 4)
        assert attrs["distance_min"] == coords["channel_number"].min() == d1
        assert attrs["distance_max"] == coords["channel_number"].max() == d2

    def test_no_arrays_in_attrs(self, quantx_v2_das_patch):
        """
        Ensure that the attributes are not arrays.
        Originally, attrs like time_min can be arrays with empty shapes.
        """
        for key, val in quantx_v2_das_patch.attrs.items():
            assert not isinstance(val, np.ndarray)


class TestIsQuantXV2:
    """Tests for function to determine if a file is an QuantXV2 file."""

    def test_format_and_version(self, quantx_v2_example_path):
        """Ensure version two is recognized."""
        format, version = QuantXV2().get_format(quantx_v2_example_path)
        assert format == QuantXV2.name
        assert version == QuantXV2.version

    def test_not_QuantXV2_not_df5(self, dummy_text_file):
        """Test for not even a hdf5 file."""
        parser = QuantXV2()
        assert not parser.get_format(dummy_text_file)
        assert not parser.get_format(dummy_text_file.parent)

    def test_hdf5file_not_quantx(self, generic_hdf5):
        """Assert that the generic hdf5 file is not a QuantXV2."""
        parser = QuantXV2()
        assert not parser.get_format(generic_hdf5)


class TestScanQuantXV2:
    """Tests for scanning QuantXV2 file."""

    def test_scanning(self, quantx_v2_example_path):
        """Tests for getting summary info from QuantXV2 data."""
        parser = QuantXV2()
        out = parser.scan(quantx_v2_example_path)
        assert isinstance(out, list)
        assert len(out) == 1
        assert isinstance(out[0], PatchFileSummary)

        data = out[0]
        assert data.file_format == parser.name
        assert data.file_version == parser.version



"""
Utilities for plotting with matplotlib.
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import dascore as dc
from dascore.constants import ONE_BILLION
from dascore.utils.time import to_number


def _get_dim_label(patch, dim):
    """Create a label for the given dimension, including units if defined."""
    attrs = patch.attrs
    maybe_units = attrs.get(f"{dim}_units")
    unit_str = f"({maybe_units})" if maybe_units else ""
    return str(dim).capitalize() + unit_str


def _get_cmap(cmap):
    """Return a color map from a colormap or string."""
    if isinstance(cmap, str):  # get color map if a string was passed
        cmap = plt.get_cmap(cmap)
    return cmap


def _get_ax(ax):
    """Get an axis if ax is None"""
    if ax is None:
        _, ax = plt.subplots(1)
    return ax


def _get_format_string(time_array):
    """Determine the best format string."""
    default_formatter = ("%Y-", "%m-", "%dT", "%H:", "%M:", "%S.%f")
    td = time_array.max() - time_array.min()
    dt = dc.to_datetime64(td / dc.to_timedelta64(1))
    dt_str = str(dt).replace(":", "-").replace("T", "-").split("-")
    base = dc.to_datetime64(0)
    base_str = str(base).replace(":", "-").replace("T", "-").split("-")
    diffs = np.array([x != y for x, y in zip(dt_str, base_str)])
    ind_min = np.argmax(diffs)
    fmter = "".join(default_formatter[ind_min:])
    return fmter


def _get_extents(dims_r, coords):
    """Get the extents used for each dimension."""
    # need to reverse dims since extent is [left, right, bottom, top]
    # and we want first dim to go from top to bottom
    lims = {x: [] for x in dims_r}
    for dim in dims_r:
        array = coords[dim]
        lims[dim] += [array.min(), array.max()]
    # special handling for time
    if "time" in coords:  # convert dates to float in seconds
        denom = ONE_BILLION
        time_min = to_number(lims["time"][0]) / denom
        time_max = to_number(lims["time"][1]) / denom
        lims["time"] = [time_min, time_max]
    out = [x for dim in dims_r for x in lims[dim]]
    return out


def _strip_labels(labels, redundants=("0", ":", "T", ".")):
    """Strip all the trailing zeros from labels."""
    ar = np.array([list(x) for x in labels])
    redundant = (np.isin(ar, redundants)).all(axis=0)
    ind2keep = np.argmax(np.cumsum((~redundant).astype(int))) + 1
    return labels.str[:ind2keep].str.rstrip(".").str.rstrip(":").str.rstrip("T")


def _get_labels(ticks, time_fmt, dt_ns):
    """Get sensible labels for plots."""
    ticks_ns = dc.to_datetime64(ticks).astype(np.int64)
    out = np.round(ticks_ns / dt_ns).astype(int) * dt_ns
    labels = pd.to_datetime(out.astype("datetime64[ns]")).strftime(time_fmt)
    return labels


def _format_time_axis(ax, patch, dims_r, time_fmt, extents=None):
    """
    Function to handle formatting time axis for image-type plots.

    Tries to snape all axis labels to "nice" values and adds reference
    start time.
    """
    time = patch.coords["time"]
    if not time_fmt:
        time_fmt = _get_format_string(time)
    approx_dt = ((time[1:] - time[:-1]) / dc.to_timedelta64(1)).mean()
    dt_ns = np.round(approx_dt * ONE_BILLION, 6).astype(int)
    # determine which axis is x, tell mpl it's a date, get date axis
    axis_name = "x" if dims_r[0] == "time" else "y"
    # getattr(ax, f"{axis_name}axis_date")()
    ticks = getattr(ax, f"get_{axis_name}ticks")()
    labels = _get_labels(ticks, time_fmt, dt_ns)
    trimmed_labels = _strip_labels(labels)
    getattr(ax, f"set_{axis_name}ticks")(ticks, trimmed_labels, rotation=-20)
    # next add time label
    _add_time_axis_label(ax, patch, dims_r, dt_ns)
    # reset limits as these can get messed up when using custom labels.
    if extents is not None:
        ax.set_xlim(extents[0], extents[1])
        ax.set_ylim(extents[2], extents[3])


def _add_time_axis_label(ax, patch, dims_r, dt_ns):
    """Add a time axis label to plot."""
    # set time label with start/end time of trace
    start, end = patch.attrs["time_min"], patch.attrs["time_max"]
    if pd.isnull(start) or pd.isnull(end):
        return  # nothing to do if no start/end times in attrs
    # round start label to within 1 dt.
    start_ns = start.astype(int)
    new_start_ns = np.round(start_ns / dt_ns).astype(int) * dt_ns
    ser = pd.Series(str(new_start_ns.astype("datetime64[ns]")))
    new_start = _strip_labels(ser).iloc[0]
    x_or_y = ["x", "y"][dims_r.index("time")]
    time_label = f"Time\n (start: {new_start})"
    getattr(ax, f"set_{x_or_y}label")(time_label)



"""
Tests for wiggle plots.
"""
import matplotlib.pyplot as plt


class TestWiggle:
    """Tests for wiggle plot"""

    def test_returns_axes(self, random_patch):
        """Call waterfall plot, return."""
        # modify patch to include line at start
        data = random_patch.data
        data[:100, :100] = 2.0  # create an origin block for testing axis line up
        data[:100, -100:] = -2.0  #
        out = random_patch.new(data=data)
        ax = out.viz.wiggle()
        # check labels
        assert random_patch.dims[0].capitalize() in ax.get_ylabel()
        assert random_patch.dims[1].capitalize() in ax.get_xlabel()
        assert isinstance(ax, plt.Axes)

    def test_example(self):
        """test the example from the docs"""
        import dascore as dc

        patch = dc.examples.sin_wave_patch(
            sample_rate=1000,
            frequency=[200, 10],
            channel_count=2,
        )
        _ = patch.viz.wiggle()



"""
Core module for reading and writing DASDAE format.
"""
import contextlib
from typing import Union

import pandas as pd

import dascore as dc
from dascore.constants import SpoolType, path_types
from dascore.core.schema import PatchFileSummary
from dascore.io.core import FiberIO
from dascore.utils.hdf5 import HDFPatchIndexManager, NodeError, open_hdf5_file
from dascore.utils.patch import get_default_patch_name

from .utils import (
    _get_contents_from_patch_groups,
    _read_patch,
    _save_patch,
    _write_meta,
)


class DASDAEV1(FiberIO):
    """
    Provides IO support for the DASDAE format version 1.

    DASDAE format is loosely based on the Adaptable Seismic Data Format (ASDF)
    which uses hdf5. The hdf5 structure is the following:

    /root
    /root.attrs
        __format__ = "DASDAE"
        __DASDAE_version__ = '1'  # version str
    /root/waveforms/
        DAS__{net}__{sta}__{tag}__{start}__{end}
            data   # patch data array
            data.attrs
            _coords_{coord_name}  # each coordinate array is saved here
        DAS__{net}__{sta}__{tag}__{start}__{end}.attrs
            _attrs_{attr_nme}  # each patch attribute
            _dims  # a str of 'dim1, dim2, dim3'
    """

    name = "DASDAE"
    preferred_extensions = ("h5", "hdf5")
    version = "1"

    def write(self, spool: SpoolType, path: path_types, index=False, **kwargs):
        """
        Write a collection of patches to a DASDAE file.

        Parameters
        ----------
        spool:
            A collection of patches or a spool (same thing).
        path
            The path to the file.
        index
            If True, create an index for the file. This slows down
            writing but will make reading/scanning the file more efficient.
            This is recommended for files with many patches and not recommended
            for files with few patches.
        """
        # write out patches
        with open_hdf5_file(path, mode="a") as h5:
            _write_meta(h5, self.version)
            # get an iterable of patches and save them
            patches = [spool] if isinstance(spool, dc.Patch) else spool
            # create new node called waveforms, else suppress error if it
            # already exists.
            with contextlib.suppress(NodeError):
                h5.create_group(h5.root, "waveforms")
            waveforms = h5.get_node("/waveforms")
            # write new patches to file
            for patch in patches:
                _save_patch(patch, waveforms, h5)
        # if the index exists, or its creation requested, update it.
        indexer = HDFPatchIndexManager(path)
        if index or indexer.has_index:
            df = self._get_patch_summary(patches)
            indexer.write_update(df)

    def _get_patch_summary(self, patches) -> pd.DataFrame:
        """Get a patch summary to put into index."""
        df = (
            dc.scan_to_df(patches)
            .assign(
                path=[f"waveforms/{get_default_patch_name(x)}" for x in patches],
                file_format=self.name,
                file_version=self.version,
            )
            .dropna(subset=["time_min", "time_max", "distance_min", "distance_max"])
        )
        return df

    def get_format(self, path) -> Union[tuple[str, str], bool]:
        """Return the format from a dasdae file."""
        with open_hdf5_file(path, mode="r") as fi:
            is_dasdae, version = False, ""  # NOQA
            with contextlib.suppress(KeyError):
                is_dasdae = fi.root._v_attrs["__format__"] == "DASDAE"
                dasdae_file_version = fi.root._v_attrs["__DASDAE_version__"]
            if is_dasdae:
                return (self.name, dasdae_file_version)
            return False

    def read(self, path, **kwargs) -> SpoolType:
        """
        Read a dascore file.
        """
        patches = []
        with open_hdf5_file(path, mode="r") as fi:
            try:
                waveform_group = fi.root["/waveforms"]
            except KeyError:
                return dc.spool([])
            for patch_group in waveform_group:
                patches.append(_read_patch(patch_group, **kwargs))
        return dc.spool(patches)

    def scan(self, path):
        """
        Get the patch info from the file.

        First we check if the file is using an index. If so, the index is
        returned. Otherwise, iterate over each waveform group and assemble
        content information.

        Parameters
        ----------
        Path
            A path to the file.
        """
        indexer = HDFPatchIndexManager(path)
        if indexer.has_index:
            # We need to change the path back to the file rather than internal
            # HDF5 path so it works with FileSpool and such.
            records = indexer.get_index().assign(path=str(path)).to_dict("records")
            return [PatchFileSummary(**x) for x in records]
        else:
            with open_hdf5_file(path) as h5:
                file_format = self.name
                version = h5.root._v_attrs.__DASDAE_version__
                return _get_contents_from_patch_groups(path, version, file_format)

    def index(self, path):
        """Index the dasdae file."""
        indexer = HDFPatchIndexManager(path)
        if not indexer.has_index:
            df = dc.scan_to_df(path)
            indexer.write_update(df)
